from cryptography.fernet import Fernet
from typing import Iterator, List, Optional, Dict
from typing import TYPE_CHECKING

from aim.storage.encoding import encode_path, decode_path
from aim.storage.types import BLOB
from aim.sdk.repo import ContainerConfig

if TYPE_CHECKING:
    from aim.storage.types import AimObjectPath
    from aim.storage.prefixview import PrefixView
    from aim.sdk.repo import Repo


def generate_resource_path(prefix_view: 'PrefixView', additional_path: 'AimObjectPath') -> str:
    prefix_path = decode_path(prefix_view.prefix)
    encoded_path = encode_path((*prefix_path, *additional_path))
    return encoded_path.hex()


class URIService:
    SEPARATOR = '__'

    def __init__(self, repo: 'Repo'):
        self.repo = repo
        self.container_persistent_pool = dict()

    @classmethod
    def generate_uri(cls, repo: 'Repo', run_name: str, sub_name: str, resource_path: str = None) -> str:
        encryptor = Fernet(key=repo.encryption_key)
        to_be_encrypted = f'{run_name}{URIService.SEPARATOR}{sub_name}'
        if resource_path:
            to_be_encrypted = f'{to_be_encrypted}{URIService.SEPARATOR}{resource_path}'

        return encryptor.encrypt(to_be_encrypted.encode()).decode()

    @classmethod
    def decode_uri(cls, repo: 'Repo', uri: str) -> List[Optional[str]]:
        decryptor = Fernet(key=repo.encryption_key)
        decrypted_uri = decryptor.decrypt(uri.encode()).decode()
        result = decrypted_uri.split(URIService.SEPARATOR, maxsplit=2)
        assert len(result) <= 3
        # extend result length to 3, fill with [None]s
        result.extend([None] * (3 - len(result)))

        return result

    def request_batch(self, uri_batch: List[str]) -> Iterator[Dict[str, bytes]]:
        for uri in uri_batch:
            run_name, sub_name, resource_path = self.decode_uri(self.repo, uri)
            container = self._get_container(run_name, sub_name)
            resource_path = decode_path(bytes.fromhex(resource_path))

            # TODO: [MV] change to some other implementation of view when available
            #  which won't collect in case of custom objects
            data = container.tree().subtree(resource_path).collect()
            if isinstance(data, BLOB):
                data = data.load()
            yield {uri: data}

        # clear container pool
        self.container_persistent_pool.clear()

    def _get_container(self, run_name: str, sub_name: str):
        config = ContainerConfig(run_name, sub_name, read_only=True)
        container = self.container_persistent_pool.get(config)

        if not container:
            if sub_name == 'meta':
                container = self.repo.request(sub_name, run_name, from_union=True, read_only=True)
            else:
                container = self.repo.request(sub_name, run_name, read_only=True)

            self.container_persistent_pool[config] = container

        return container



import os
import time


def timing(iter_count=5):
    def inner(f):
        def wrapper(*args, **kwargs):
            ts = time.time()
            for _ in range(iter_count):
                f(*args, **kwargs)
            te = time.time()
            return (te-ts)/iter_count
        return wrapper
    return inner


def get_baseline_filename():
    import performance_tests
    performance_tests_path = os.path.dirname(performance_tests.__file__)
    baseline_filename = os.path.join(performance_tests_path, 'BASELINE')
    if os.environ.get('AIM_PERFORMANCE_TESTS_BASELINE'):
        # for local performance testing
        baseline_filename = os.environ['AIM_PERFORMANCE_TESTS_BASELINE']

    return baseline_filename


def get_baseline(test_name):
    filename = get_baseline_filename()
    if not os.path.exists(filename):
        return None

    with open(filename, 'r') as f:
        for line in f:
            if test_name in line:
                return float(line.split()[1])

    return None


def write_baseline(test_name, exec_time):
    filename = get_baseline_filename()

    with open(filename, 'a+') as f:
        f.write(f'{test_name} {exec_time}\n')



import os
import shutil
import logging

from collections import defaultdict
from contextlib import contextmanager
from enum import Enum
from cachetools.func import ttl_cache
from typing import Dict, Tuple, Iterator, NamedTuple, Optional, List, Set
from weakref import WeakValueDictionary

from aim.ext.sshfs.utils import mount_remote_repo, unmount_remote_repo
from aim.ext.task_queue.queue import TaskQueue
from aim.ext.cleanup import AutoClean
from aim.ext.transport.client import Client

from aim.sdk.configs import get_aim_repo_name, AIM_ENABLE_TRACKING_THREAD
from aim.sdk.errors import RepoIntegrityError
from aim.sdk.run import Run
from aim.sdk.utils import search_aim_repo, clean_repo_path
from aim.sdk.sequence_collection import QuerySequenceCollection, QueryRunSequenceCollection
from aim.sdk.sequence import Sequence
from aim.sdk.types import QueryReportMode
from aim.sdk.data_version import DATA_VERSION
from aim.sdk.remote_repo_proxy import RemoteRepoProxy
from aim.sdk.lock_manager import LockManager, RunLock

from aim.storage.locking import SoftFileLock
from aim.storage.container import Container
from aim.storage.rockscontainer import RocksContainer, LockableRocksContainer
from aim.storage.union import RocksUnionContainer
from aim.storage.treeviewproxy import ProxyTree
from aim.storage.lock_proxy import ProxyLock

from aim.storage.structured.db import DB
from aim.storage.structured.proxy import StructuredRunProxy

logger = logging.getLogger(__name__)


class ContainerConfig(NamedTuple):
    name: str
    sub: Optional[str]
    read_only: bool


class RepoStatus(Enum):
    MISSING = 1
    UPDATE_REQUIRED = 2
    PATCH_REQUIRED = 3
    UPDATED = 4


def _get_tracking_queue():
    if os.getenv(AIM_ENABLE_TRACKING_THREAD, False):
        return TaskQueue('metric_tracking', max_backlog=10_000_000)  # single thread task queue for Run.track
    return None


class RepoAutoClean(AutoClean):
    PRIORITY = 30

    def __init__(self, instance: 'Repo') -> None:
        """
        Prepare the `Repo` for automatic cleanup.

        Args:
            instance: The `Repo` instance to be cleaned up.
        """
        super().__init__(instance)
        self.root_path = instance.root_path
        self._client = instance._client
        self._mount_root = instance._mount_root

    def _close(self) -> None:
        """Close the `Repo` and unmount the remote repository."""
        if self._client:
            self._client._heartbeat_sender.stop()
            self._client.disconnect()
        if self._mount_root:
            logger.debug(f'Unmounting remote repository at {self._mount_root}')
            unmount_remote_repo(self.root_path, self._mount_root)


# TODO make this api thread-safe
class Repo:
    """Aim repository object.

    Provides methods for  repositories creation/opening/cleanup.
    Provides APIs for accessing Runs.
    Provides API for querying Runs/Metrics based on a given expression.

    Args:
        path (str): Path to Aim repository.
        read_only (:obj:`bool`, optional): Flag for opening Repo in readonly mode. False by default.
        init (:obj:`bool`, optional): Flag used to initialize new Repo. False by default.
            Recommended to use ``aim init`` command instead.
    """
    _pool = WeakValueDictionary()  # TODO: take read only into account

    tracking_queue = _get_tracking_queue()

    def __init__(self, path: str, *, read_only: Optional[bool] = None, init: Optional[bool] = False):
        if read_only is not None:
            raise NotImplementedError

        self._resources = None
        self.read_only = read_only
        self._mount_root = None
        self._client: Client = None
        self._lock_manager: LockManager = None
        if path.startswith('ssh://'):
            self._mount_root, self.root_path = mount_remote_repo(path)
        elif self.is_remote_path(path):
            remote_path = path.replace('aim://', '')
            self._client = Client(remote_path)
            self._remote_repo_proxy = RemoteRepoProxy(self._client)
            self.root_path = remote_path
        else:
            self.root_path = path
        self.path = os.path.join(self.root_path, get_aim_repo_name())

        if init:
            os.makedirs(self.path, exist_ok=True)
            os.makedirs(os.path.join(self.path, 'locks'), exist_ok=True)
        if not self.is_remote_repo and not os.path.exists(self.path):
            if self._mount_root:
                unmount_remote_repo(self.root_path, self._mount_root)
            raise RuntimeError(f'Cannot find repository \'{self.path}\'. Please init first.')

        self.container_pool: Dict[ContainerConfig, Container] = WeakValueDictionary()
        self.persistent_pool: Dict[ContainerConfig, Container] = dict()
        self.container_view_pool: Dict[ContainerConfig, Container] = WeakValueDictionary()

        self._run_props_cache_hint = None
        self._encryption_key = None
        self.structured_db = None

        if not self.is_remote_repo:
            self._lock_manager = LockManager(self.path)
            self._sdb_lock_path = os.path.join(self.path, 'locks', 'structured_db_lock')
            self._sdb_lock = SoftFileLock(self._sdb_lock_path, timeout=5 * 60)  # timeout after 5 minutes

            status = self.check_repo_status(self.root_path)
            self.structured_db = DB.from_path(self.path)
            if init or status == RepoStatus.PATCH_REQUIRED:
                with self._sdb_lock:
                    self.structured_db.run_upgrades()
                    with open(os.path.join(self.path, 'VERSION'), 'w') as version_fh:
                        version_fh.write('.'.join(map(str, DATA_VERSION)) + '\n')

        self._resources = RepoAutoClean(self)

    @property
    def meta_tree(self):
        return self.request_tree('meta', read_only=True, from_union=True).subtree('meta')

    def __repr__(self) -> str:
        return f'<Repo#{hash(self)} path={self.path} read_only={self.read_only}>'

    def __hash__(self) -> int:
        return hash(self.path)

    def __eq__(self, o: 'Repo') -> bool:
        return self.path == o.path

    @classmethod
    def default_repo_path(cls) -> str:
        repo_path, found = search_aim_repo(os.path.curdir)
        if not found:
            repo_path = os.getcwd()
        return repo_path

    @classmethod
    def default_repo(cls, init: bool = False):
        """Named constructor for default repository.

        Searches nearest `.aim` directory from current directory to roo directory.
        If not found, return Repo for current directory.

        Args:
            init (:obj:`bool`, optional): Flag used to initialize new Repo. False by default.
                Recommended to use `aim init` command instead.
        Returns:
            :obj:`Repo` object.
        """
        return cls.from_path(cls.default_repo_path(), init=init)

    @classmethod
    def from_path(cls, path: str, read_only: bool = None, init: bool = False):
        """Named constructor for Repo for given path.

        Arguments:
            path (str): Path to Aim repository.
            read_only (:obj:`bool`, optional): Flag for opening Repo in readonly mode. False by default.
            init (:obj:`bool`, optional): Flag used to initialize new Repo. False by default.
                Recommended to use ``aim init`` command instead.
        Returns:
            :obj:`Repo` object.
        """
        if not path.startswith('ssh://') and not cls.is_remote_path(path):
            path = clean_repo_path(path)
        repo = cls._pool.get(path)
        if repo is None:
            repo = Repo(path, read_only=read_only, init=init)
            cls._pool[path] = repo
        return repo

    @classmethod
    def exists(cls, path: str) -> bool:
        """Check Aim repository existence.

        Args:
            path (str): Path to Aim repository.
        Returns:
            True if repository exists, False otherwise.
        """
        path = clean_repo_path(path)
        aim_repo_path = os.path.join(path, get_aim_repo_name())
        return os.path.exists(aim_repo_path)

    @classmethod
    def rm(cls, path: str):
        """Remove Aim repository.

        Args:
            path (str): Path to Aim repository.
        """
        path = clean_repo_path(path)
        repo = cls._pool.get(path)
        if repo is not None:
            del cls._pool[path]
        aim_repo_path = os.path.join(path, get_aim_repo_name())
        shutil.rmtree(aim_repo_path)

    @classmethod
    def check_repo_status(cls, path: str) -> RepoStatus:
        if cls.is_remote_path(path):
            return RepoStatus.UPDATED
        if not cls.exists(path):
            return RepoStatus.MISSING
        repo_version = cls.get_version(path)
        current_version = DATA_VERSION
        if repo_version < current_version:
            if repo_version[0] < current_version[0]:
                return RepoStatus.UPDATE_REQUIRED
            else:
                return RepoStatus.PATCH_REQUIRED
        return RepoStatus.UPDATED

    @classmethod
    def get_version(cls, path: str):
        path = clean_repo_path(path)
        version_file_path = os.path.join(path, get_aim_repo_name(), 'VERSION')
        if os.path.exists(version_file_path):
            with open(version_file_path, 'r') as version_fh:
                version_tp = version_fh.read().strip().split('.')
                return tuple(map(int, version_tp))
        return 0, 0  # old Aim repos

    @classmethod
    def is_remote_path(cls, path: str):
        return path.startswith('aim://')

    def _get_container(
            self, name: str, read_only: bool, from_union: bool = False
    ) -> Container:
        if self.read_only and not read_only:
            raise ValueError('Repo is read-only')

        container_config = ContainerConfig(name, None, read_only=read_only)
        container = self.container_pool.get(container_config)
        if container is None:
            path = os.path.join(self.path, name)
            if from_union:
                container = RocksUnionContainer(path, read_only=read_only)
                self.persistent_pool[container_config] = container
            else:
                container = RocksContainer(path, read_only=read_only)
            self.container_pool[container_config] = container

        return container

    def _get_index_tree(self, name: str, timeout: int):
        if not self.is_remote_repo:
            return self._get_index_container(name, timeout).tree()
        else:
            return ProxyTree(self._client, name, '', read_only=False, index=True, timeout=timeout)

    def _get_index_container(self, name: str, timeout: int) -> Container:
        if self.read_only:
            raise ValueError('Repo is read-only')

        name = name + '/index'
        container_config = ContainerConfig(name, None, read_only=True)
        container = self.container_pool.get(container_config)
        if container is None:
            path = os.path.join(self.path, name)
            container = LockableRocksContainer(path, read_only=False, timeout=timeout)
            self.container_pool[container_config] = container

        return container

    def request_tree(
        self,
        name: str,
        sub: str = None,
        *,
        read_only: bool,
        from_union: bool = False  # TODO maybe = True by default
    ):
        if not self.is_remote_repo:
            return self.request(name, sub, read_only=read_only, from_union=from_union).tree()
        else:
            return ProxyTree(self._client, name, sub, read_only, from_union)

    def request(
            self,
            name: str,
            sub: str = None,
            *,
            read_only: bool,
            from_union: bool = False  # TODO maybe = True by default
    ):

        container_config = ContainerConfig(name, sub, read_only)
        container_view = self.container_view_pool.get(container_config)
        if container_view is None:
            if read_only:
                if from_union:
                    path = name
                else:
                    assert sub is not None
                    path = os.path.join(name, 'chunks', sub)
                container = self._get_container(path, read_only=True, from_union=from_union)
            else:
                assert sub is not None
                path = os.path.join(name, 'chunks', sub)
                container = self._get_container(path, read_only=False, from_union=False)

            container_view = container
            self.container_view_pool[container_config] = container_view

        return container_view

    def request_props(self, hash_: str, read_only: bool):
        if self.is_remote_repo:
            return StructuredRunProxy(self._client, hash_, read_only)

        assert self.structured_db
        _props = None

        if self.run_props_cache_hint:
            _props = self.structured_db.caches[self.run_props_cache_hint][hash_]
        if not _props:
            _props = self.structured_db.find_run(hash_)
            if not _props:
                if read_only:
                    raise RepoIntegrityError(f'Missing props for Run {hash_}')
                else:
                    with self._sdb_lock:
                        _props = self.structured_db.create_run(hash_)
            if self.run_props_cache_hint:
                self.structured_db.caches[self.run_props_cache_hint][hash_] = _props

        return _props

    def request_run_lock(self, hash_: str, timeout: int = 10) -> 'RunLock':
        if self.is_remote_repo:
            return ProxyLock(self._client, hash_)
        assert self._lock_manager
        return self._lock_manager.get_run_lock(hash_, timeout=timeout)

    def iter_runs(self) -> Iterator['Run']:
        """Iterate over Repo runs.

        Yields:
            next :obj:`Run` in readonly mode .
        """
        self.meta_tree.preload()
        for run_name in self.meta_tree.subtree('chunks').keys():
            yield Run(run_name, repo=self, read_only=True)

    def iter_runs_from_cache(self, offset: str = None) -> Iterator['Run']:
        db = self.structured_db
        cache = db.caches.get('runs_cache')
        if cache:
            run_names = cache.keys()
            try:
                offset_idx = run_names.index(offset) + 1
            except ValueError:
                offset_idx = 0
            for run_name in run_names[offset_idx:]:
                yield Run(run_name, repo=self, read_only=True)
        else:
            raise StopIteration

    def run_exists(self, run_hash: str) -> bool:
        return run_hash in self._all_run_hashes()

    @ttl_cache(ttl=0.5)
    def _all_run_hashes(self) -> Set[str]:
        if self.is_remote_repo:
            return set(self._remote_repo_proxy.list_all_runs())
        else:
            chunks_dir = os.path.join(self.path, 'meta', 'chunks')
            if os.path.exists(chunks_dir):
                return set(os.listdir(chunks_dir))
            else:
                return set()

    def list_all_runs(self) -> List[str]:
        return list(self._all_run_hashes())

    def _active_run_hashes(self) -> Set[str]:
        if self.is_remote_repo:
            return set(self._remote_repo_proxy.list_active_runs())
        else:
            chunks_dir = os.path.join(self.path, 'meta', 'progress')
            if os.path.exists(chunks_dir):
                return set(os.listdir(chunks_dir))
            else:
                return set()

    def list_active_runs(self) -> List[str]:
        return list(self._active_run_hashes())

    def total_runs_count(self) -> int:
        db = self.structured_db
        if db:
            cache = db.caches.get('runs_cache')
        else:
            cache = None
        if cache:
            return len(cache.keys())
        else:
            self.meta_tree.preload()
            return len(list(self.meta_tree.subtree('chunks').keys()))

    def get_run(self, run_hash: str) -> Optional['Run']:
        """Get run if exists.

        Args:
            run_hash (str): Run hash.
        Returns:
            :obj:`Run` object if hash is found in repository. `None` otherwise.
        """
        # TODO: [MV] optimize existence check for run
        if run_hash is None or run_hash not in self.meta_tree.subtree('chunks').keys():
            return None
        else:
            return Run(run_hash, repo=self, read_only=True)

    def query_runs(self,
                   query: str = '',
                   paginated: bool = False,
                   offset: str = None,
                   report_mode: QueryReportMode = QueryReportMode.PROGRESS_BAR) -> QueryRunSequenceCollection:
        """Get runs satisfying query expression.

        Args:
             query (:obj:`str`, optional): query expression.
                If not specified, query results will include all runs.
             paginated (:obj:`bool`, optional): query results pagination flag. False if not specified.
             offset (:obj:`str`, optional): `hash` of Run to skip to.
             report_mode(:obj:`QueryReportMode`, optional): indicates report mode
                (0: DISABLED, 1: PROGRESS BAR, 2: PROGRESS TUPLE). QueryReportMode.PROGRESS_BAR if not specified.
        Returns:
            :obj:`SequenceCollection`: Iterable for runs/metrics matching query expression.
        """
        self._prepare_runs_cache()
        return QueryRunSequenceCollection(self, Sequence, query, paginated, offset, report_mode)

    def delete_run(self, run_hash: str) -> bool:
        """Delete Run data from aim repository

        This action removes run data permanently and cannot be reverted.
        If you want to archive run but keep it's data use `repo.get_run(run_hash).archived = True`.

        Args:
            run_hash (:obj:`str`): Run to be deleted.

        Returns:
            True if run deleted successfully, False otherwise.
        """
        try:
            self._delete_run(run_hash)
            return True
        except Exception as e:
            logger.warning(f'Error while trying to delete run \'{run_hash}\'. {str(e)}.')
            return False

    def delete_runs(self, run_hashes: List[str]) -> Tuple[bool, List[str]]:
        """Delete multiple Runs data from aim repository

        This action removes runs data permanently and cannot be reverted.
        If you want to archive run but keep it's data use `repo.get_run(run_hash).archived = True`.

        Args:
            run_hashes (:obj:`str`): list of Runs to be deleted.

        Returns:
            (True, []) if all runs deleted successfully, (False, :obj:`list`) with list of remaining runs otherwise.
        """
        remaining_runs = []
        for run_hash in run_hashes:
            try:
                self._delete_run(run_hash)
            except Exception as e:
                logger.warning(f'Error while trying to delete run \'{run_hash}\'. {str(e)}.')
                remaining_runs.append(run_hash)

        if remaining_runs:
            return False, remaining_runs
        else:
            return True, []

    def copy_runs(self, run_hashes: List[str], dest_repo: 'Repo') -> Tuple[bool, List[str]]:
        """Copy multiple Runs data from current aim repository to destination aim repository

        Args:
            run_hashes (:obj:`str`): list of Runs to be copied.
            dest_repo (:obj:`Repo`): destination Repo instance to copy Runs

        Returns:
            (True, []) if all runs were copied successfully,
            (False, :obj:`list`) with list of remaining runs otherwise.
        """
        remaining_runs = []
        for run_hash in run_hashes:
            try:
                self._copy_run(run_hash, dest_repo)
            except Exception as e:
                logger.warning(f'Error while trying to copy run \'{run_hash}\'. {str(e)}.')
                remaining_runs.append(run_hash)

        if remaining_runs:
            return False, remaining_runs
        else:
            return True, []

    def move_runs(self, run_hashes: List[str], dest_repo: 'Repo') -> Tuple[bool, List[str]]:
        """Move multiple Runs data from current aim repository to destination aim repository

        Args:
            run_hashes (:obj:`str`): list of Runs to be moved.
            dest_repo (:obj:`Repo`): destination Repo instance to move Runs

        Returns:
            (True, []) if all runs were moved successfully,
            (False, :obj:`list`) with list of remaining runs otherwise.
        """
        remaining_runs = []
        for run_hash in run_hashes:
            try:
                self._copy_run(run_hash, dest_repo)
                self._delete_run(run_hash)
            except Exception as e:
                logger.warning(f'Error while trying to move run \'{run_hash}\'. {str(e)}.')
                remaining_runs.append(run_hash)

        if remaining_runs:
            return False, remaining_runs
        else:
            return True, []

    def query_metrics(self,
                      query: str = '',
                      report_mode: QueryReportMode = QueryReportMode.PROGRESS_BAR) -> QuerySequenceCollection:
        """Get metrics satisfying query expression.

        Args:
             query (str): query expression.
             report_mode(:obj:`QueryReportMode`, optional): indicates report mode
                (0: DISABLED, 1: PROGRESS BAR, 2: PROGRESS TUPLE). QueryReportMode.PROGRESS_BAR if not specified.
        Returns:
            :obj:`SequenceCollection`: Iterable for metrics matching query expression.
        """
        self._prepare_runs_cache()
        from aim.sdk.sequences.metric import Metric
        return QuerySequenceCollection(repo=self, seq_cls=Metric, query=query, report_mode=report_mode)

    def query_images(self,
                     query: str = '',
                     report_mode: QueryReportMode = QueryReportMode.PROGRESS_BAR) -> QuerySequenceCollection:
        """Get image collections satisfying query expression.

        Args:
             query (str): query expression.
             report_mode(:obj:`QueryReportMode`, optional): indicates report mode
                (0: DISABLED, 1: PROGRESS BAR, 2: PROGRESS TUPLE). QueryReportMode.PROGRESS_BAR if not specified.
        Returns:
            :obj:`SequenceCollection`: Iterable for image sequences matching query expression.
        """
        self._prepare_runs_cache()
        from aim.sdk.sequences.image_sequence import Images
        return QuerySequenceCollection(repo=self, seq_cls=Images, query=query, report_mode=report_mode)

    def query_audios(self,
                     query: str = '',
                     report_mode: QueryReportMode = QueryReportMode.PROGRESS_BAR) -> QuerySequenceCollection:
        """Get audio collections satisfying query expression.

        Args:
             query (str): query expression.
             report_mode(:obj:`QueryReportMode`, optional): indicates report mode
                (0: DISABLED, 1: PROGRESS BAR, 2: PROGRESS TUPLE). QueryReportMode.PROGRESS_BAR if not specified.
        Returns:
            :obj:`SequenceCollection`: Iterable for audio sequences matching query expression.
        """
        self._prepare_runs_cache()
        from aim.sdk.sequences.audio_sequence import Audios
        return QuerySequenceCollection(repo=self, seq_cls=Audios, query=query, report_mode=report_mode)

    def query_figure_objects(self,
                             query: str = '',
                             report_mode: QueryReportMode = QueryReportMode.PROGRESS_BAR) -> QuerySequenceCollection:
        """Get Figures collections satisfying query expression.

        Args:
             query (str): query expression.
             report_mode(:obj:`QueryReportMode`, optional): indicates report mode
                (0: DISABLED, 1: PROGRESS BAR, 2: PROGRESS TUPLE). QueryReportMode.PROGRESS_BAR if not specified.
        Returns:
            :obj:`SequenceCollection`: Iterable for Figure sequences matching query expression.
        """
        self._prepare_runs_cache()
        from aim.sdk.sequences.figure_sequence import Figures
        return QuerySequenceCollection(repo=self, seq_cls=Figures, query=query, report_mode=report_mode)

    def query_distributions(self,
                            query: str = '',
                            report_mode: QueryReportMode = QueryReportMode.PROGRESS_BAR) -> QuerySequenceCollection:
        """Get distribution collections satisfying query expression.

        Args:
             query (str): query expression.
             report_mode(:obj:`QueryReportMode`, optional): indicates report mode
                (0: DISABLED, 1: PROGRESS BAR, 2: PROGRESS TUPLE). QueryReportMode.PROGRESS_BAR if not specified.
        Returns:
            :obj:`SequenceCollection`: Iterable for distribution sequences matching query expression.
        """
        self._prepare_runs_cache()
        from aim.sdk.sequences.distribution_sequence import Distributions
        return QuerySequenceCollection(repo=self, seq_cls=Distributions, query=query, report_mode=report_mode)

    def query_texts(self,
                    query: str = '',
                    report_mode: QueryReportMode = QueryReportMode.PROGRESS_BAR) -> QuerySequenceCollection:
        """Get text collections satisfying query expression.

        Args:
             query (str): query expression.
             report_mode(:obj:`QueryReportMode`, optional): indicates report mode
                (0: DISABLED, 1: PROGRESS BAR, 2: PROGRESS TUPLE). QueryReportMode.PROGRESS_BAR if not specified.
        Returns:
            :obj:`SequenceCollection`: Iterable for text sequences matching query expression.
        """
        self._prepare_runs_cache()
        from aim.sdk.sequences.text_sequence import Texts
        return QuerySequenceCollection(repo=self, seq_cls=Texts, query=query, report_mode=report_mode)

    @property
    def run_props_cache_hint(self):
        return self._run_props_cache_hint

    @run_props_cache_hint.setter
    def run_props_cache_hint(self, cache: str):
        self._run_props_cache_hint = cache

    @property
    def encryption_key(self):
        from cryptography.fernet import Fernet

        if self._encryption_key:
            return self._encryption_key

        encryption_key_path = os.path.join(self.path, 'ENCRYPTION_KEY')
        if not os.path.exists(encryption_key_path):
            with open(encryption_key_path, 'w') as key_fp:
                encryption_key = Fernet.generate_key().decode()
                key_fp.write(encryption_key + '\n')
        else:
            with open(encryption_key_path, 'r') as key_fp:
                encryption_key = key_fp.readline()

        self._encryption_key = encryption_key

        return encryption_key

    def _get_meta_tree(self):
        return self.request_tree(
            'meta', read_only=True, from_union=True
        ).subtree('meta')

    @staticmethod
    def available_sequence_types():
        return Sequence.registry.keys()

    @staticmethod
    def validate_sequence_types(sequence_types: Tuple[str, ...]):
        for seq_name in sequence_types:
            seq_cls = Sequence.registry.get(seq_name, None)
            if seq_cls is None or not issubclass(seq_cls, Sequence):
                raise ValueError(f'\'{seq_name}\' is not a valid Sequence')

    def collect_sequence_info(self, sequence_types: Tuple[str, ...]) -> Dict[str, Dict[str, list]]:
        """Utility function for getting sequence names and contexts for all runs by given sequence types.

        Args:
            sequence_types (:obj:`tuple[str]`, optional): Sequence types to get tracked sequence names/contexts for.
            Defaults to 'metric'.

        Returns:
            :obj:`dict`: Tree of sequences and their contexts groupped by sequence type.
        """
        meta_tree = self._get_meta_tree()
        sequence_traces = {}
        if isinstance(sequence_types, str):
            sequence_types = (sequence_types,)
        for seq_type in sequence_types:
            seq_cls = Sequence.registry.get(seq_type, None)
            if seq_cls is None:
                raise ValueError(f'\'{seq_type}\' is not a valid Sequence')
            assert issubclass(seq_cls, Sequence)
            dtypes = seq_cls.allowed_dtypes()
            dtype_traces = set()
            for dtype in dtypes:
                try:
                    dtype_trace_tree = meta_tree.collect(('traces_types', dtype))
                    for ctx_id, seqs in dtype_trace_tree.items():
                        for seq_name in seqs.keys():
                            dtype_traces.add((ctx_id, seq_name))
                except KeyError:
                    pass
            if 'float' in dtypes:  # old sequences without dtype set are considered float sequences
                try:
                    dtype_trace_tree = meta_tree.collect('traces')
                    for ctx_id, seqs in dtype_trace_tree.items():
                        for seq_name in seqs.keys():
                            dtype_traces.add((ctx_id, seq_name))
                except KeyError:
                    pass
            traces_info = defaultdict(list)
            for ctx_id, seq_name in dtype_traces:
                traces_info[seq_name].append(meta_tree['contexts', ctx_id])
            sequence_traces[seq_type] = traces_info
        return sequence_traces

    def collect_params_info(self) -> dict:
        """Utility function for getting run meta-parameters.

        Returns:
            :obj:`dict`: All runs meta-parameters.
        """
        meta_tree = self._get_meta_tree()
        try:
            return meta_tree.collect('attrs', strict=False)
        except KeyError:
            return {}

    def _prepare_runs_cache(self):
        if self.is_remote_repo:
            return

        db = self.structured_db
        cache_name = 'runs_cache'
        db.invalidate_cache(cache_name)
        db.init_cache(cache_name, db.runs, lambda run: run.hash)
        self.run_props_cache_hint = cache_name

    def _delete_run(self, run_hash):
        # check run lock info. in progress runs can't be deleted
        if self._lock_manager.get_run_lock_info(run_hash).locked:
            raise RuntimeError(f'Cannot delete Run \'{run_hash}\'. Run is locked.')

        with self.structured_db:  # rollback db entity delete if subsequent actions fail.
            # remove database entry
            self.structured_db.delete_run(run_hash)

            # remove data from index container
            index_tree = self._get_index_container('meta', timeout=0).tree()
            del index_tree.subtree(('meta', 'chunks'))[run_hash]

            # delete rocksdb containers data
            sub_dirs = ('chunks', 'progress', 'locks')
            for sub_dir in sub_dirs:
                meta_path = os.path.join(self.path, 'meta', sub_dir, run_hash)
                if os.path.isfile(meta_path):
                    os.remove(meta_path)
                else:
                    shutil.rmtree(meta_path, ignore_errors=True)
                seqs_path = os.path.join(self.path, 'seqs', sub_dir, run_hash)
                if os.path.isfile(seqs_path):
                    os.remove(seqs_path)
                else:
                    shutil.rmtree(seqs_path, ignore_errors=True)

    def _copy_run(self, run_hash, dest_repo):
        # check run lock info. in progress runs can't be copied
        if self._lock_manager.get_run_lock_info(run_hash).locked:
            raise RuntimeError(f'Cannot copy Run \'{run_hash}\'. Run is locked.')

        with dest_repo.structured_db:  # rollback destination db entity if subsequent actions fail.
            # copy run structured data
            source_structured_run = self.structured_db.find_run(run_hash)
            # create destination structured run db instance, set experiment and archived state
            dest_structured_run = dest_repo.structured_db.create_run(run_hash, source_structured_run.created_at)
            dest_structured_run.experiment = source_structured_run.experiment
            dest_structured_run.archived = source_structured_run.archived
            # create and add to the destination run source tags
            for source_tag in source_structured_run.tags_obj:
                try:
                    dest_tag = dest_repo.structured_db.create_tag(source_tag.name)
                    dest_tag.color = source_tag.color
                    dest_tag.description = source_tag.description
                except ValueError:
                    pass  # if the tag already exists in destination db no need to do anything
                dest_structured_run.add_tag(source_tag.name)

            # copy run meta tree
            source_meta_run_tree = self.request_tree(
                'meta', run_hash, read_only=True, from_union=True
            ).subtree('meta').subtree('chunks').subtree(run_hash)
            dest_meta_run_tree = dest_repo.request_tree(
                'meta', run_hash, read_only=False, from_union=True
            ).subtree('meta').subtree('chunks').subtree(run_hash)
            dest_meta_run_tree[...] = source_meta_run_tree[...]
            dest_index = dest_repo._get_index_tree('meta', timeout=0).view(())
            dest_meta_run_tree.finalize(index=dest_index)

            # copy run series tree
            source_series_run_tree = self.request_tree(
                'seqs', run_hash, read_only=True
            ).subtree('seqs').subtree('chunks').subtree(run_hash)
            dest_series_run_tree = dest_repo.request_tree(
                'seqs', run_hash, read_only=False
            ).subtree('seqs').subtree('chunks').subtree(run_hash)
            dest_series_run_tree[...] = source_series_run_tree[...]

    def close(self):
        if self._resources is None:
            return
        self._resources.close()

    @property
    def is_remote_repo(self):
        return self._client is not None

    @contextmanager
    def atomic_track(self, queue_id):
        if self.is_remote_repo:
            self._client.start_instructions_batch()
        yield
        if self.is_remote_repo:
            self._client.flush_instructions_batch(queue_id)



"""run end time

Revision ID: 3c4f22db7a46
Revises: 9ba30ab3b2b4
Create Date: 2021-09-01 21:19:00.391203

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '3c4f22db7a46'
down_revision = '9ba30ab3b2b4'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('run', sa.Column('finalized_at', sa.DateTime(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('run', 'finalized_at')
    # ### end Alembic commands ###



from typing import Dict, Any, Union, Optional, Tuple


class Trace(object):
    def __init__(self, repo, metric, name: str, context: Optional[list] = None):
        self.repo = repo
        self.metric = metric
        self.name = name
        self.data = []
        self.tmp_data = {}  # used for storing temporary data for x_axis metric values
        self.current_x_axis_value = None  # used for checking ordering of x_axis metric values
        self.alignment = None  # used for storing alignment flags when x_axis is aligned by custom metric
        self.slice = None
        self._num_records = None
        self.context = None  # type: Optional[Dict[str, Union[str, Any]]]
        self.hashable_context = None
        if context is not None:
            self.context = {
                k: v for (k, v) in context
            }
            self.hashable_context = tuple(sorted(self.context.items()))

    def __repr__(self):
        return str(self.context)

    def __len__(self):
        return self.num_records

    def read_records(self, indices: Optional[Union[int, Tuple[int, ...],
                                                   slice]] = None):
        storage = self.metric.get_storage()
        records_iter = storage.read_records(self.name, indices,
                                            self.context)
        return records_iter



from aim.web.api.dashboards.models import Dashboard
from aim.web.api.dashboard_apps.models import ExploreState


def dashboard_response_serializer(dashboard_object, session):
    if not isinstance(dashboard_object, Dashboard):
        return None

    app = session.query(ExploreState)\
        .filter(ExploreState.dashboard_id == dashboard_object.uuid, ExploreState.is_archived == False)\
        .first()

    response = {
        'id': dashboard_object.uuid,
        'name': dashboard_object.name,
        'description': dashboard_object.description,
        'app_id': app.uuid if app else None,
        'app_type': app.type if app else None,
        'updated_at': dashboard_object.updated_at,
        'created_at': dashboard_object.created_at
    }

    return response



import datetime

from typing import List, TYPE_CHECKING

import aim.ext.transport.remote_router_pb2 as router_rpc
import aim.ext.transport.remote_router_pb2_grpc as router_pb2_grpc

from aim.ext.transport.message_utils import build_exception

if TYPE_CHECKING:
    from aim.ext.transport.worker import RemoteWorker


class RemoteRouterServicer(router_pb2_grpc.RemoteRouterServiceServicer):
    client_heartbeat_pool = dict()

    def __init__(self, worker_pool: List['RemoteWorker']):
        self.worker_pool = worker_pool

    def new_client(self, client_uri):
        sorted_workers_pool = sorted(self.worker_pool, key=lambda w: w.client_count)

        worker = sorted_workers_pool[0]
        worker.add_client(client_uri)
        return worker

    def get_version(self, request: router_rpc.VersionRequest, _context):
        from aim.__version__ import __version__ as aim_version

        return router_rpc.VersionResponse(version=aim_version,
                                          status=router_rpc.VersionResponse.Status.OK)

    def client_heartbeat(self, request: router_rpc.HeartbeatRequest, _context):
        try:
            client_uri = request.client_uri
            self.client_heartbeat_pool[client_uri] = datetime.datetime.now().timestamp()
            return router_rpc.HeartbeatResponse(status=router_rpc.HeartbeatResponse.Status.OK)
        except Exception as e:
            return router_rpc.HeartbeatResponse(
                status=router_rpc.HeartbeatRequest.Status.ERROR,
                exception=build_exception(e),
            )

    def connect(self, request: router_rpc.ConnectRequest, _context):
        try:
            worker = self.new_client(request.client_uri)
            return router_rpc.ConnectResponse(port=worker.port,
                                              worker_index=worker.index,
                                              status=router_rpc.ConnectResponse.Status.OK)
        except Exception as e:
            return router_rpc.ConnectResponse(status=router_rpc.ConnectResponse.Status.ERROR,
                                              exception=build_exception(e))

    def reconnect(self, request: router_rpc.ReconnectRequest, _context):
        try:
            client_uri = request.client_uri
            for worker in self.worker_pool:
                if client_uri in worker.clients:
                    worker.restart()
                    return router_rpc.ReconnectResponse(port=worker.port,
                                                        status=router_rpc.ReconnectResponse.Status.OK)
            # if client wasn't found in the list of clients of any worker fallback to connection logic
            worker = self.new_client(client_uri)
            return router_rpc.ReconnectResponse(port=worker.port,
                                                status=router_rpc.ReconnectResponse.Status.OK)
        except Exception as e:
            return router_rpc.ReconnectResponse(status=router_rpc.ReconnectResponse.Status.ERROR,
                                                exception=build_exception(e))

    def disconnect(self, request: router_rpc.DisconnectRequest, _context):
        try:
            client_uri = request.client_uri
            for worker in self.worker_pool:
                worker.remove_client(client_uri)

            return router_rpc.DisconnectResponse(status=router_rpc.DisconnectResponse.Status.OK)
        except Exception as e:
            return router_rpc.DisconnectResponse(status=router_rpc.DisconnectResponse.Status.ERROR,
                                                 exception=build_exception(e))



import logging

logger = logging.getLogger(__name__)

DEFAULT_MSG_TEMPLATE = 'This functionality will be removed in'


def python_version_deprecation_check():
    import sys
    version_info = sys.version_info
    if version_info.major == 3 and version_info.minor == 6:
        deprecation_warning(remove_version='3.16',
                            msg='Python 3.6 has reached EOL. Aim support for Python 3.6 is deprecated!',
                            remove_msg_template='Python 3.6 support will be dropped in')


def deprecation_warning(*, remove_version: str, msg: str, remove_msg_template: str = DEFAULT_MSG_TEMPLATE):
    logger.warning(msg)
    logger.warning(f'{remove_msg_template} Aim version {remove_version}')



import grpc
import os
import traceback
from base64 import b64decode

PUBLIC_KEY = os.getenv('AIM_AUTH_PUBLIC_KEY')
AUDIENCE = os.getenv('AIM_AUTH_AUDIENCE')
ALGO = os.getenv('AIM_AUTH_ALGO', 'RS256')
wrapped_public_key = f"""-----BEGIN PUBLIC KEY-----\n{PUBLIC_KEY}\n-----END PUBLIC KEY-----"""

def load_basic_users():
    users = {}
    if os.getenv('AIM_AUTH_HTPASSWD_FILE'):
        with open(os.getenv('AIM_AUTH_HTPASSWD_FILE')) as f:
            for record in filter(None, map(lambda x: x.strip(), f.readlines())):
                username, parts = record.split(':')
                salt = parts.split('/')[0]
                users[username.encode('utf-8')] = (salt.encode('utf-8'), parts.encode('utf-8'))
    return users

users = load_basic_users()


class ServerAuthInterceptor(grpc.ServerInterceptor):
    def __init__(self):
        def deny(_, context):
            context.abort(grpc.StatusCode.UNAUTHENTICATED, 'Invalid Credentials')
        self._deny = grpc.unary_unary_rpc_method_handler(deny)

    def intercept_service(self, continuation, handler_call_details):
        import jwt
        import bcrypt
        metas = handler_call_details.invocation_metadata
        for metadatum in metas:
            if metadatum.key == 'authorization':
                try:
                    if metadatum.value.startswith('Bearer '):
                        token = metadatum.value.split(maxsplit=1)[1]
                        jwt.decode(token, wrapped_public_key, audience=AUDIENCE, algorithms=ALGO)
                    elif metadatum.value.startswith('Basic '):
                        cr = metadatum.value.split(maxsplit=1)[1].encode('utf-8')
                        username, password = b64decode(cr).split(b':', maxsplit=1)
                        if username in users:
                            salt, parts = users[username]
                            if bcrypt.hashpw(password, salt) != parts:
                                return self._deny
                        else:
                            return self._deny
                    return continuation(handler_call_details)
                except Exception as e:    
                    traceback.print_exc()
                    return self._deny
        return self._deny


# distutils: language = c++
# cython: language_level = 3
# cython: wraparound = False
# cython: boundscheck = False
# cython: cdivision=True
# cython: nonecheck=False

from libcpp.vector cimport vector
from libcpp.pair cimport pair


cpdef enum:
    PATH_SENTINEL_CODE = 0xfe

ctypedef long long int64
ctypedef unsigned long long uint64

cpdef bytes encode_int64_big_endian(int64 value)

cpdef bytes encode_int64(int64 value)

cpdef int64 decode_int64_big_endian(const unsigned char* b, int offset = *) nogil

cpdef int64 decode_int64(const unsigned char* buffer, int offset = *) nogil

cpdef vector[pair[int64, int64]] split_path(
    const unsigned char* buffer,
    int64 length
) nogil

cpdef list decode_path(bytes buffer)

cpdef bytes encode_double(double value)

cpdef double decode_double(const unsigned char* bytes, int offset = *) nogil

cpdef bytes encode_utf_8_str(str value)

cpdef str decode_utf_8_str(bytes buffer)



from typing import TYPE_CHECKING

from aim.ext.transport.message_utils import pack_args
from aim.ext.transport.remote_resource import RemoteResourceAutoClean

from aim.storage.locking import RunLock
from aim.storage.treeutils import encode_tree

if TYPE_CHECKING:
    from aim.ext.transport.client import Client


class ProxyLockAutoClean(RemoteResourceAutoClean):
    PRIORITY = 45


class ProxyLock(RunLock):
    def __init__(self, client: 'Client',
                 run_hash: str):
        self._resources: ProxyLockAutoClean = None

        self._rpc_client = client
        self._hash = run_hash

        kwargs = {
            'run_hash': run_hash,
        }
        self.init_args = pack_args(encode_tree(kwargs))
        self.resource_type = 'Lock'
        handler = self._rpc_client.get_resource_handler(self, self.resource_type, args=self.init_args)

        self._resources = ProxyLockAutoClean(self)
        self._resources.rpc_client = client
        self._resources.handler = handler
        self._handler = handler

    def lock(self, force: bool = False) -> None:
        return self._rpc_client.run_instruction(self._hash, self._handler, 'lock', (force,))

    def release(self, force: bool = False) -> None:
        return self._rpc_client.run_instruction(self._hash, self._handler, 'release', (force,))



import logging
import os.path

from PIL import (
    Image as PILImage,
    ImageSequence as PILImageSequence
)

from io import BytesIO
from itertools import chain, repeat
from typing import List
import numpy as np

from aim.sdk.num_utils import inst_has_typename
from aim.storage.object import CustomObject
from aim.storage.types import BLOB

logger = logging.getLogger(__name__)


@CustomObject.alias('aim.image')
class Image(CustomObject):
    """Image object used to store image objects in Aim repository...

    Args:
         image (:obj:): pillow `Image` object or `torch.Tensor` or `numpy.array` used to construct `aim.Image`.
         caption (:obj:`str`, optional): Optional image caption. '' by default.
         format (:obj: `str`, optional): Parameter for PIL's .save() method. 'png' by default.
         quality (:obj: `int`, optional): Parameter for PIL's .save() method. 85 by default.
         optimize (:obj: `bool`, optional): Parameter for PIL's .save() method. False by default.

         For more information on the format, quality and optimize parameters, refer to PIL documentation.

    Example of params to reduce quality of the image:
        format='jpeg',
        optimize=True,
        quality=85
    """

    DEFAULT_IMG_FORMAT = 'png'
    FLAG_WARN_RGBA_RGB = False
    AIM_NAME = 'aim.image'

    def __init__(self, image, caption: str = '', format=None, quality=90, optimize=False):
        super().__init__()

        params = {
            'format': format,
            'quality': quality,
            'optimize': optimize
        }

        if isinstance(image, str):
            self._from_file_path(image, params)
        elif inst_has_typename(image, ['PIL', 'Image']):
            self._from_pil_image(image, params)
        elif inst_has_typename(image, ['torch', 'Tensor']):
            self._from_torch_tensor(image, params)
        elif inst_has_typename(image, ['tensorflow', 'Tensor']):
            self._from_tf_tensor(image, params)
        elif inst_has_typename(image, ['numpy', 'array']):
            self._from_numpy_array(image, params)
        elif inst_has_typename(image, ['Figure', 'matplotlib', 'figure']):
            self._from_matplotlib_figure(image, params)
        else:
            raise TypeError(f'Cannot convert to aim.Image. Unsupported type {type(image)}.')
        self.caption = caption

    @property
    def caption(self) -> str:
        """Image caption, set by user.

            :getter: Returns image caption.
            :setter: Sets image caption.
            :type: string
        """
        return self.storage['caption']

    @caption.setter
    def caption(self, value: str):
        self.storage['caption'] = value

    @property
    def format(self) -> str:
        """Stored image format.

            :getter: Returns image format.
            :type: string
        """
        return self.storage['format']

    @property
    def width(self):
        """Stored image width.

            :getter: Returns image width.
            :type: string
        """
        return self.storage['width']

    @property
    def height(self):
        """Stored image height.

            :getter: Returns image height.
            :type: string
        """
        return self.storage['height']

    @property
    def size(self):
        """Stored image size.

            :getter: Returns image (width, height) pair.
            :type: string
        """
        return self.storage['width'], self.storage['height']

    def to_pil_image(self) -> PILImage.Image:
        """Method to convert aim.Image to pillow Image"""
        pil_img = PILImage.open(BytesIO(bytes(self.storage['data'])))
        assert pil_img.size == self.size
        return pil_img

    def json(self):
        """Dump image metadata to a dict"""
        return {
            'caption': self.caption,
            'format': self.format,
            'width': self.width,
            'height': self.height,
        }

    def _from_pil_image(self, pil_image: PILImage.Image, params):
        assert isinstance(pil_image, PILImage.Image)
        img_container = BytesIO()

        if not params['format']:
            params['format'] = pil_image.format or self.DEFAULT_IMG_FORMAT
        else:
            # normalize img format
            img_format = params['format'].lower()
            if img_format == 'jpg':
                # PIL doesn't support 'jpg' key
                params['format'] = 'jpeg'
            params['format'] = img_format

        if getattr(pil_image, "n_frames", 1) > 1:
            # is animated
            frames = PILImageSequence.all_frames(pil_image)
            params.update(
                dict(
                    save_all=True,
                    append_images=frames[1:],
                )
            )
            frames[0].save(
                img_container,
                **params
            )
        else:
            try:
                pil_image.save(img_container, **params)
            except OSError as exc:
                # The best way to approach this problem is to prepare PIL Image object before hitting this method.
                # This block only handles case where RGBA/P/LA/PA mode is mandated to save in RGB
                # PIL won't do that automatically, so we have to convert image to RGB before saving it.
                # In addition - make transparency "white" before conversion otherwise it will be black.
                if pil_image.mode not in ('RGBA', 'LA', 'PA', 'P'):
                    raise exc
                elif not Image.FLAG_WARN_RGBA_RGB:
                    logger.warning(f'Failed to save the image due to the following error: {exc}')
                    logger.warning(f'Attempting to convert mode "{pil_image.mode}" to "RGB"')
                    Image.FLAG_WARN_RGBA_RGB = True

                alpha = pil_image.convert('RGBA').split()[-1]  # Get only alpha
                background = PILImage.new('RGBA', pil_image.size, (255, 255, 255, 255))
                background.paste(pil_image, mask=alpha)
                pil_image = background.convert('RGB')

                # Retry
                pil_image.save(img_container, **params)

        self.storage['data'] = BLOB(data=img_container.getvalue())
        self.storage['source'] = 'PIL.Image'
        self.storage['mode'] = pil_image.mode
        self.storage['format'] = params['format']
        self.storage['width'], self.storage['height'] = pil_image.size

    def _from_file_path(self, file_path, params):
        if not os.path.isfile(file_path):
            raise ValueError('Invalid image file path.')

        return self._from_pil_image(PILImage.open(file_path), params)

    def _from_numpy_array(self, array: np.ndarray, params):
        if array.ndim not in {2, 3}:
            raise ValueError('Cannot convert to aim.Image. array must have 2/3-D shape.')

        if array.ndim == 3 and array.shape[2] == 1:  # greyscale
            pil_image = PILImage.fromarray(array[:, :, 0])
        else:
            pil_image = PILImage.fromarray(array)
        self._from_pil_image(pil_image, params)

    def _from_torch_tensor(self, tensor, params):
        try:
            import torch
            assert isinstance(tensor, torch.Tensor)
        except (ImportError, AssertionError):
            raise ValueError('Cannot convert from torch.Tensor')

        if tensor.ndim not in {2, 3}:
            raise ValueError('Cannot convert to aim.Image. Tensor must have 2/3-D shape.')
        if tensor.is_floating_point():
            tensor = tensor.mul(255).byte()
        array: np.ndarray = tensor.cpu().numpy()

        if array.ndim == 3:
            channels = array.shape[0]
            if channels == 1:
                pil_image = PILImage.fromarray(array[0, :, :])
            else:
                # reverse order of channels: c,h,w => h,w,c
                pil_image = PILImage.fromarray(np.transpose(array, (1, 2, 0)))
        else:
            pil_image = PILImage.fromarray(array)
        self._from_pil_image(pil_image, params)

    def _from_tf_tensor(self, tensor, params):
        try:
            import tensorflow as tf
            assert isinstance(tensor, tf.Tensor)
        except (ImportError, AssertionError):
            raise ValueError('Cannot convert from tf.Tensor')

        if tensor.ndim not in {2, 3}:
            raise ValueError('Cannot convert to aim.Image. Tensor must have 2/3-D shape.')

        if tensor.dtype.is_floating:
            tensor = tf.cast(tf.math.scalar_mul(255.0, tensor), tf.uint8)
        array: np.ndarray = tensor.numpy()

        if array.ndim == 3 and array.shape[2] == 1:  # greyscale
            pil_image = PILImage.fromarray(array[:, :, 0])
        else:
            pil_image = PILImage.fromarray(array)
        self._from_pil_image(pil_image, params)

    def _from_matplotlib_figure(self, figure, params):
        try:
            assert hasattr(figure, 'savefig')
        except AssertionError:
            raise ValueError('Cannot convert from matplotlib figure.')

        buffer = BytesIO()
        figure.savefig(buffer)
        buffer.seek(0)

        return self._from_pil_image(PILImage.open(buffer), params)

    def __eq__(self, other):
        if not isinstance(other, Image):
            return False

        props = ['mode', 'format', 'width', 'height']
        for p in props:
            if self.storage[p] != other.storage[p]:
                return False

        return self.storage['data'].load() == other.storage['data'].load()


def convert_to_aim_image_list(images, labels=None) -> List[Image]:
    aim_images = []
    if labels is not None:
        labels_it = chain(labels, repeat(''))
    else:
        labels_it = repeat('')
    for img, lbl in zip(images, labels_it):
        aim_img = Image(img, lbl if isinstance(lbl, str) else str(lbl.item()))
        aim_images.append(aim_img)
    return aim_images



import time
import os

here = os.path.abspath(os.path.dirname(__file__))

version_file = f'{here}/VERSION'

__version__ = None
with open(version_file) as vf:
    __version__ = vf.read().strip()

_this_year = time.strftime("%Y")
__author__ = "Gev Sogomonian, Gor Arakelyan et al."
__author_email__ = "community@aimhub.io"
__license__ = "Apache-2.0"
__copyright__ = f"2019-{_this_year}, {__author__}."
__homepage__ = "https://github.com/aimhubio/aim"
__docs_url__ = "https://github.com/aimhubio/aim"
# this has to be simple string, see: https://github.com/pypa/twine/issues/522
__docs__ = (
    "Aim is a lightweight library for ML researchers to store and retrieve AI metadata."
)
__long_docs__ = """
"""

__all__ = ["__author__", "__author_email__", "__copyright__", "__docs__", "__homepage__", "__license__", "__version__"]



from aim.storage.object import CustomObject
import hub


@CustomObject.alias('hub.dataset')
class HubDataset(CustomObject):
    AIM_NAME = 'hub.dataset'

    def __init__(self, dataset: hub.Dataset):
        super().__init__()
        self.storage['dataset'] = {
            'source': 'hub',
            'meta': self._get_ds_meta(dataset)
        }

    def _get_ds_meta(self, ds: hub.Dataset):
        return {
            "path": ds.path,
            "version": ds.meta.version,
            "info": ds.info._info,  # Info might contain keys such as "description" and "title"
            "tensors": {k: self._tensor_meta(v) for k, v in ds.tensors.items()},
            "num_samples": len(ds),
        }

    def _tensor_meta(self, tensor: hub.Tensor):
        meta = tensor.meta
        return {
            "name": tensor.key,
            "num_samples": len(tensor),
            "htype": tensor.htype,
            "dtype": str(tensor.dtype) if tensor.dtype else None,
            "compression_type": "sample_compression"
            if meta.sample_compression
            else ("chunk_compression" if meta.chunk_compression else None),
            "compression_format": meta.sample_compression or meta.chunk_compression,
            "info": tensor.info._info,
        }



import pytest

from tests.base import TestBase
from tests.utils import is_package_installed


class TestHubDatasetIntegration(TestBase):
    @pytest.mark.skipif(not is_package_installed('hub'), reason="'hub' is not installed. skipping.")
    def test_dataset_as_run_param(self):
        import hub

        from aim.sdk.objects.plugins.hub_dataset import HubDataset
        from aim.sdk import Run

        # create dataset object
        ds = hub.dataset('hub://activeloop/cifar100-test')

        # log dataset metadata
        run = Run(system_tracking_interval=None)
        run['hub_ds'] = HubDataset(ds)

        # get dataset metadata
        ds_object = run['hub_ds']
        ds_dict = run.get('hub_ds', resolve_objects=True)

        self.assertTrue(isinstance(ds_object, HubDataset))
        self.assertTrue(isinstance(ds_dict, dict))
        self.assertIn('meta', ds_dict['dataset'].keys())
        self.assertIn('source', ds_dict['dataset'].keys())



import numpy as np

from typing import Any, Iterator, List, Tuple, Union


class ArrayView:
    """Array of homogeneous elements with sparse indices.
    Interface for working with array as a non-sparse array is available for cases
    when index values are not important.
    """

    def __iter__(self) -> Iterator[Any]:
        ...

    def keys(self) -> Iterator[int]:
        """Return sparse indices iterator.

        Yields:
             Array's next sparse index.
        """
        ...

    def indices(self) -> Iterator[int]:
        """Return sparse indices iterator.

        Yields:
             Array's next sparse index.
        """
        ...

    def values(self) -> Iterator[Any]:
        """Return values iterator.

        Yields:
             Array's next value.
        """
        ...

    def items(self) -> Iterator[Tuple[int, Any]]:
        """Return items iterator.

        Yields:
            Tuple of array's next sparse index and value.
        """
        ...

    def __len__(self) -> int:
        ...

    def __getitem__(
        self,
        idx: Union[int, slice]
    ):
        ...

    # TODO implement append

    def __setitem__(
        self,
        idx: int,
        val: Any
    ):
        ...

    def sparse_list(self) -> Tuple[List[int], List[Any]]:
        """Get sparse indices and values as :obj:`list`s.
        """
        ...

    def indices_list(self) -> List[int]:
        """Get sparse indices as a :obj:`list`.
        """
        ...

    def values_list(self) -> List[Any]:
        """Get values as a :obj:`list`.
        """
        ...

    def sparse_numpy(self) -> Tuple[np.ndarray, np.ndarray]:
        """Get sparse indices and values as numpy arrays.
        """
        ...

    def indices_numpy(self) -> np.ndarray:
        """Get sparse indices as numpy array.
        """
        ...

    def values_numpy(self) -> np.ndarray:
        """Get values as numpy array.
        """
        ...

    def tolist(self) -> List[Any]:
        """Convert to values list"""
        ...

    def first(self) -> Tuple[int, Any]:
        """First index and value of the array.
        """
        ...

    def first_idx(self) -> int:
        """First index of the array.
        """
        ...

    def first_value(self) -> Any:
        """First value of the array.
        """
        ...

    def last(self) -> Tuple[int, Any]:
        """Last index and value of the array.
        """
        ...

    def last_idx(self) -> int:
        """Last index of the array.
        """
        ...

    def last_value(self) -> Any:
        """Last value of the array.
        """
        ...



import uuid

from tests.base import PrefilledDataApiTestBase


class TestStructuredRunApi(PrefilledDataApiTestBase):
    def setUp(self) -> None:
        super().setUp()

        self.test_id = str(uuid.uuid4())
        with self.repo.structured_db:
            for idx, run in zip(range(10), self.repo.iter_runs()):
                exp_name = f'Experiment 1 {self.test_id}' if idx < 5 else f'Experiment 2 {self.test_id}'

                run.name = f'Run number {idx + 1} {self.test_id}'
                run.experiment = exp_name
                if idx < 3:
                    run.add_tag(f'first runs {self.test_id}')
                elif 3 <= idx < 7:
                    run.add_tag(f'first runs {self.test_id}')
                    run.add_tag(f'last runs {self.test_id}')
                else:
                    run.add_tag(f'last runs {self.test_id}')

    # runs API
    def test_set_run_experiment_api(self):
        matching_runs = self.repo.structured_db.search_runs(f'Run number 3 {self.test_id}')
        run = next(iter(matching_runs))
        self.assertEqual(f'Experiment 1 {self.test_id}', run.experiment)

        client = self.client
        # set existing experiment
        resp = client.put(f'/api/runs/{run.hash}', json={'experiment': f'Experiment 2 {self.test_id}'})
        self.assertEqual(200, resp.status_code)
        self.assertEqual(f'Experiment 2 {self.test_id}', run.experiment)

        # set non-existing experiment (create new)
        resp = client.put(f'/api/runs/{run.hash}', json={'experiment': f'New experiment {self.test_id}'})
        self.assertEqual(200, resp.status_code)
        self.assertEqual(f'New experiment {self.test_id}', run.experiment)

    def test_add_remove_tag_api(self):
        matching_runs = self.repo.structured_db.search_runs(f'Run number 5 {self.test_id}')
        run = next(iter(matching_runs))
        tags = [tag for tag in run.tags_obj]
        self.assertEqual(2, len(run.tags_obj))

        self.assertListEqual([f'first runs {self.test_id}', f'last runs {self.test_id}'], run.tags)

        client = self.client
        # remove tag # 1
        resp = client.delete(f'/api/runs/{run.hash}/tags/{tags[0].uuid}')
        self.assertEqual(200, resp.status_code)
        self.assertEqual(1, len(run.tags))
        self.assertListEqual([f'last runs {self.test_id}'], run.tags)

        # add new tag
        resp = client.post(f'/api/runs/{run.hash}/tags/new', json={'tag_name': f'new tag {self.test_id}'})
        self.assertEqual(200, resp.status_code)
        self.assertEqual(2, len(run.tags))
        self.assertListEqual([f'last runs {self.test_id}', f'new tag {self.test_id}'], run.tags)

    def test_update_run_name_description(self):
        matching_runs = self.repo.structured_db.search_runs('Run number 3')
        run = next(iter(matching_runs))
        client = self.client
        resp = client.put(f'/api/runs/{run.hash}', json={'description': f'long text {self.test_id}',
                                                         'name': f'best run {self.test_id}'})
        self.assertEqual(200, resp.status_code)
        self.assertEqual(f'best run {self.test_id}', run.name)
        self.assertEqual(f'long text {self.test_id}', run.description)

    # tags API
    def test_list_tags_api(self):
        client = self.client
        response = client.get('/api/tags')
        self.assertEqual(200, response.status_code)
        data = response.json()
        test_tags = [tag for tag in data if tag['name'].endswith(self.test_id)]
        self.assertEqual(2, len(test_tags))

    def test_search_tags_api(self):
        client = self.client
        response = client.get('/api/tags/search', params={'q': self.test_id})
        self.assertEqual(200, response.status_code)
        data = response.json()
        self.assertEqual(2, len(data))

        response = client.get('/api/tags/search', params={'q': f'last runs {self.test_id}'})
        self.assertEqual(200, response.status_code)
        data = response.json()
        self.assertEqual(1, len(data))

    def test_get_tag_api(self):
        client = self.client
        response = client.get('/api/tags/search', params={'q': f'last runs {self.test_id}'})
        self.assertEqual(200, response.status_code)
        data = response.json()

        tag_uuid = data[0]['id']
        response = client.get(f'/api/tags/{tag_uuid}')
        self.assertEqual(200, response.status_code)
        data = response.json()
        self.assertEqual(f'last runs {self.test_id}', data['name'])
        self.assertEqual(None, data['color'])
        self.assertFalse(data['archived'])

    def test_create_tag_api(self):
        client = self.client
        response = client.post('/api/tags/', json={'name': f'my awesome tag {self.test_id}'})
        self.assertEqual(200, response.status_code)
        data = response.json()
        tag_uuid = data['id']
        new_tag = self.repo.structured_db.find_tag(tag_uuid)
        self.assertEqual(f'my awesome tag {self.test_id}', new_tag.name)

    def test_update_tag_props_api(self):
        tag = next(iter(self.repo.structured_db.tags()))
        self.assertEqual(None, tag.color)
        client = self.client
        response = client.put(f'/api/tags/{tag.uuid}/', json={'name': 'my awesome tag',
                                                              'color': '#FFFFFF',
                                                              'description': 'new description'})
        self.assertEqual(200, response.status_code)
        self.assertEqual('my awesome tag', tag.name)
        self.assertEqual('#FFFFFF', tag.color)
        self.assertEqual('new description', tag.description)

    def test_get_tag_runs_api(self):
        client = self.client
        response = client.get('/api/tags/search', params={'q': 'last runs'})
        self.assertEqual(200, response.status_code)
        data = response.json()

        tag_uuid = data[0]['id']
        response = client.get(f'/api/tags/{tag_uuid}/runs/')
        self.assertEqual(200, response.status_code)
        data = response.json()
        run_names = {run['name'] for run in data['runs']}
        expected_run_names = {f'Run number {i} {self.test_id}' for i in range(4, 11)}
        self.assertSetEqual(expected_run_names, run_names)
        self.assertTrue(all('end_time' in run for run in data['runs']))

    def test_archive_tag_api(self):
        tag = next(iter(self.repo.structured_db.tags()))
        client = self.client
        response = client.put(f'/api/tags/{tag.uuid}/', json={'archived': True})
        self.assertEqual(200, response.status_code)
        self.assertTrue(tag.archived)

    def test_delete_tag_api(self):
        tag = next(iter(self.repo.structured_db.tags()))
        client = self.client
        delete_response = client.delete(f'/api/tags/{tag.uuid}/')
        self.assertEqual(200, delete_response.status_code)
        get_response = client.delete(f'/api/tags/{tag.uuid}/')
        self.assertEqual(404, get_response.status_code)

    def test_list_run_tags_with_archived_tag(self):
        tag = next(iter(self.repo.structured_db.tags()))
        self.assertTrue(all(tag in r.tags for r in tag.runs))
        client = self.client
        response = client.put(f'/api/tags/{tag.uuid}/', json={'archived': True})
        self.assertEqual(200, response.status_code)
        self.assertFalse(any(tag in r.tags for r in tag.runs))

    # experiments API
    def test_list_experiments_api(self):
        client = self.client
        response = client.get('/api/experiments')
        self.assertEqual(200, response.status_code)
        data = response.json()
        test_experiments = [exp for exp in data if exp['name'].endswith(self.test_id)]
        self.assertEqual(2, len(test_experiments))  # count default experiment
        default_experiment = next((exp for exp in data if exp['name'] == 'default'), None)
        self.assertIsNotNone(default_experiment)

    def test_search_experiments_api(self):
        client = self.client
        response = client.get('/api/experiments/search', params={'q': f'{self.test_id}'})
        self.assertEqual(200, response.status_code)
        data = response.json()
        self.assertEqual(2, len(data))

        response = client.get('/api/experiments/search', params={'q': f'Experiment 2 {self.test_id}'})
        self.assertEqual(200, response.status_code)
        data = response.json()
        self.assertEqual(1, len(data))

    def test_get_experiment_api(self):
        client = self.client
        response = client.get('/api/experiments/search', params={'q': f'Experiment 2 {self.test_id}'})
        self.assertEqual(200, response.status_code)
        data = response.json()

        exp_uuid = data[0]['id']
        response = client.get(f'/api/experiments/{exp_uuid}')
        self.assertEqual(200, response.status_code)
        data = response.json()
        self.assertEqual(f'Experiment 2 {self.test_id}', data['name'])

    def test_create_experiment_api(self):
        client = self.client
        response = client.post('/api/experiments/', json={'name': f'New experiment {self.test_id}'})
        self.assertEqual(200, response.status_code)
        data = response.json()
        exp_uuid = data['id']
        exp = self.repo.structured_db.find_experiment(exp_uuid)
        self.assertEqual(f'New experiment {self.test_id}', exp.name)

    def test_update_experiment_props_api(self):
        exp = next(iter(self.repo.structured_db.experiments()))
        client = self.client
        response = client.put(f'/api/experiments/{exp.uuid}/', json={'name': 'Updated experiment'})
        self.assertEqual(200, response.status_code)
        self.assertEqual('Updated experiment', exp.name)

    def test_get_experiment_runs_api(self):
        client = self.client
        response = client.get('/api/experiments/search', params={'q': f'Experiment 2 {self.test_id}'})
        self.assertEqual(200, response.status_code)
        data = response.json()

        exp_uuid = data[0]['id']
        response = client.get(f'/api/experiments/{exp_uuid}/runs/')
        self.assertEqual(200, response.status_code)
        data = response.json()
        run_names = {run['name'] for run in data['runs']}
        expected_run_names = {f'Run number {i} {self.test_id}' for i in range(6, 11)}
        self.assertSetEqual(expected_run_names, run_names)

    def test_get_experiment_runs_paginated_api(self):
        client = self.client
        response = client.get('/api/experiments/search', params={'q': f'Experiment 2 {self.test_id}'})
        data = response.json()

        exp_uuid = data[0]['id']
        response = client.get(f'/api/experiments/{exp_uuid}/runs/', params={'limit': 2})
        self.assertEqual(200, response.status_code)
        data = response.json()
        self.assertEqual(2, len(data['runs']))

        offset = data['runs'][-1]['run_id']

        response = client.get(f'/api/experiments/{exp_uuid}/runs/', params={'limit': 5, 'offset': offset})
        self.assertEqual(200, response.status_code)
        data = response.json()
        run_ids = {run['run_id'] for run in data['runs']}
        self.assertNotIn(offset, run_ids)
        self.assertEqual(3, len(data['runs']))

    def test_archive_experiment_with_runs(self):
        client = self.client
        response = client.get('/api/experiments/search', params={'q': f'Experiment 2 {self.test_id}'})
        self.assertEqual(200, response.status_code)
        data = response.json()
        exp_uuid = data[0]['id']

        response = client.put(f'/api/experiments/{exp_uuid}/', json={'archived': True})
        self.assertEqual(response.status_code, 400)
        error_msg = f'Cannot archive experiment \'{exp_uuid}\'. Experiment has associated runs.'
        self.assertEqual(response.json()['message'], error_msg)



from typing import Optional

from aim.sdk.run import Run
from aim.sdk.adapters.keras_mixins import TrackerKerasCallbackMetricsEpochEndMixin
from aim.ext.resource.configs import DEFAULT_SYSTEM_TRACKING_INT

try:
    from tensorflow.keras.callbacks import Callback
except ImportError:
    raise RuntimeError(
        'This contrib module requires tensorflow to be installed. '
        'Please install it with command: \n pip install tensorflow'
    )


class AimCallback(TrackerKerasCallbackMetricsEpochEndMixin, Callback):
    def __init__(self, repo: Optional[str] = None,
                 experiment: Optional[str] = None,
                 system_tracking_interval: int = DEFAULT_SYSTEM_TRACKING_INT,
                 log_system_params: bool = True,):
        super(Callback, self).__init__()

        self._system_tracking_interval = system_tracking_interval
        self._log_system_params = log_system_params

        if repo is None and experiment is None:
            self._run = Run(system_tracking_interval=self._system_tracking_interval,
                            log_system_params=self._log_system_params,)
        else:
            self._run = Run(repo=repo, experiment=experiment,
                            system_tracking_interval=self._system_tracking_interval,
                            log_system_params=self._log_system_params,)

        self._run_hash = self._run.hash
        self._repo_path = repo

    @property
    def experiment(self) -> Run:
        if not self._run:
            self._run = Run(self._run_hash,
                            repo=self._repo_path,
                            system_tracking_interval=self._system_tracking_interval,)
        return self._run

    @classmethod
    def metrics(cls, repo: Optional[str] = None,
                experiment: Optional[str] = None,
                run: Optional[Run] = None):
        # Keep `metrics` method for backward compatibility
        return cls(repo, experiment, run)

    def close(self) -> None:
        if self._run:
            self._run.close()
            del self._run
            self._run = None

    def __del__(self):
        self.close()


# Keep `AimTracker` for backward compatibility
AimTracker = AimCallback



from functools import wraps


def exception_resistant(silent: bool):
    def inner(func):
        if not silent:
            num_fails = 0
            max_fails = 6

            @wraps(func)
            def wrapper(*args, **kwargs):
                nonlocal num_fails
                func_name = func.__name__
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    num_fails += 1
                    if num_fails == 1:
                        print('Something went wrong in `{}`. The process will continue to execute.'.format(func_name))
                    if num_fails <= max_fails:
                        print('`{}`: {}'.format(func_name, e))
                    elif num_fails == max_fails + 1:
                        print('The rest of the `{}` errors are hidden.'.format(func_name))
        else:
            @wraps(func)
            def wrapper(*args, **kwargs):
                try:
                    return func(*args, **kwargs)
                except Exception:
                    pass
        return wrapper
    return inner



import pytest
import numpy as np
from parameterized import parameterized

from tests.base import PrefilledDataApiTestBase, ApiTestBase
from tests.utils import decode_encoded_tree_stream

from aim.storage.treeutils import decode_tree
from aim.sdk.run import Run


class TestRunApi(PrefilledDataApiTestBase):
    def test_search_runs_api(self):
        client = self.client

        query = self.isolated_query_patch('run["name"] == "Run # 3"')
        response = client.get('/api/runs/search/run/', params={'q': query,
                                                               'report_progress': False})
        self.assertEqual(200, response.status_code)

        decoded_response = decode_tree(decode_encoded_tree_stream(response.iter_content(chunk_size=512*1024)))
        self.assertEqual(1, len(decoded_response))
        for _, run in decoded_response.items():
            self.assertEqual(4, len(run['traces']['metric']))
            for trace in run['traces']['metric']:
                self.assertAlmostEqual(0.99, trace['last_value']['last'])

    def test_search_runs_api_paginated(self):
        client = self.client

        query = self.isolated_query_patch('run["name"] in ["Run # 2","Run # 3"]')
        response = client.get('/api/runs/search/run/', params={'q': query,
                                                               'limit': 1, 'report_progress': False})
        self.assertEqual(200, response.status_code)

        decoded_response = decode_tree(decode_encoded_tree_stream(response.iter_content(chunk_size=1024 * 1024)))
        self.assertEqual(1, len(decoded_response))

        offset = ''
        for run_hash, run in decoded_response.items():
            offset = run_hash
            self.assertEqual('Run # 3', run['props']['name'])

        query = self.isolated_query_patch('run["name"] in ["Run # 2","Run # 3"]')
        response = client.get('/api/runs/search/run/', params={'q': query,
                                                               'limit': 5,
                                                               'offset': offset,
                                                               'report_progress': False})
        self.assertEqual(200, response.status_code)

        decoded_response = decode_tree(decode_encoded_tree_stream(response.iter_content(chunk_size=1024 * 1024)))
        self.assertEqual(1, len(decoded_response))
        for run_hash, run in decoded_response.items():
            self.assertEqual('Run # 2', run['props']['name'])

    def test_search_metrics_api_default_step(self):
        client = self.client

        query = self.isolated_query_patch('run["name"] == "Run # 3"')
        response = client.get('/api/runs/search/metric/', params={'q': query,
                                                                  'report_progress': False})
        self.assertEqual(200, response.status_code)

        decoded_response = decode_tree(decode_encoded_tree_stream(response.iter_content(chunk_size=512*1024)))
        for run in decoded_response.values():
            for trace in run['traces']:
                self.assertEqual([0, 0, 50], trace['slice'])
                values = trace['values']
                dtype = values['dtype']
                shape = values['shape']
                data = values['blob']
                array = np.frombuffer(data, dtype=dtype).reshape(shape)
                self.assertEqual(50, shape)  # default 50 steps
                self.assertAlmostEqual(0.99, array.max())
                self.assertAlmostEqual(0.99, array[49])
                self.assertEqual(50, len(array))

    @parameterized.expand([
        (10,),
        (2,),
        (90,),
    ])
    def test_search_metrics_api_custom_step(self, step_count):
        client = self.client

        query = self.isolated_query_patch('run["name"] == "Run # 3"')
        response = client.get('/api/runs/search/metric/', params={'q': query,
                                                                  'p': step_count,
                                                                  'report_progress': False})
        self.assertEqual(200, response.status_code)

        decoded_response = decode_tree(decode_encoded_tree_stream(response.iter_content(chunk_size=512*1024)))
        for run in decoded_response.values():
            for trace in run['traces']:
                self.assertEqual([0, 0, step_count], trace['slice'])
                values = trace['values']
                dtype = values['dtype']
                shape = values['shape']
                data = values['blob']
                array = np.frombuffer(data, dtype=dtype).reshape(shape)
                self.assertEqual(step_count, shape)
                self.assertAlmostEqual(0.99, array.max())
                self.assertAlmostEqual(0.99, array[step_count - 1])
                self.assertEqual(step_count, len(array))

    def test_search_aligned_metrics_api(self):
        client = self.client
        run_hashes = []
        for run, _ in zip(self.repo.iter_runs(), range(2)):
            run_hashes.append(run.hash)

        response = client.post('/api/runs/search/metric/align/', json={
            'align_by': 'accuracy',
            'runs': [{
                'run_id': run_hashes[0],
                'traces': [{'name': 'loss', 'slice': [0, 0, 100], 'context': {'is_training': False}}]
            }, {
                'run_id': run_hashes[1],
                'traces': [{'name': 'loss', 'slice': [0, 0, 100], 'context': {'is_training': True, 'subset': 'train'}}]
            }]
        })
        self.assertEqual(200, response.status_code)

        decoded_response = decode_tree(decode_encoded_tree_stream(response.iter_content(chunk_size=512*1024)))
        self.assertEqual(2, len(decoded_response))
        self.assertListEqual(run_hashes, list(decoded_response.keys()))
        self.assertEqual([], decoded_response[run_hashes[1]])
        traces = decoded_response[run_hashes[0]][0]
        self.assertEqual('loss', traces['name'])
        self.assertDictEqual({'is_training': False}, traces['context'])
        self.assertEqual(99, traces['x_axis_values']['shape'])

    @pytest.mark.skip(reason="low priority. requires more investigation.")
    def test_search_aligned_metrics_api_with_wrong_context(self):
        client = self.client
        run_hashes = []
        for run, _ in zip(self.repo.iter_runs(), range(2)):
            run_hashes.append(run.hash)

        response = client.post('/api/runs/search/metric/align/', json={
            'align_by': 'accuracy',
            'runs': [{
                'run_id': run_hashes[0],
                'traces': [{'name': 'loss', 'slice': [0, 20, 1], 'context': {'is_training': True, 'subset': 'training'}}]
            }, {
                'run_id': run_hashes[1],
                'traces': [{'name': 'loss', 'slice': [0, 10, 1], 'context': {'is_training': True, 'subset': 'val'}}]
            }]
        })
        self.assertEqual(200, response.status_code)
        decoded_response = decode_tree(decode_encoded_tree_stream(response.iter_content(chunk_size=1024 * 1024)))
        self.assertEqual(2, len(decoded_response))
        for run in decoded_response.values():
            self.assertEqual([], run)

    def test_run_info_api(self):
        run = self._find_run_by_name('Run # 1')

        self.assertEqual('Run # 1', run.name)
        client = self.client
        response = client.get(f'/api/runs/{run.hash}/info/')
        self.assertEqual(200, response.status_code)

        data = response.json()
        run_params = data['params']
        self.assertEqual(1, run_params['run_index'])
        self.assertEqual(0.001, run_params['hparams']['lr'])

        run_traces_overview = data['traces']['metric']
        self.assertEqual(4, len(run_traces_overview))
        for trc_overview in run_traces_overview:
            self.assertIn(trc_overview['name'], ['loss', 'accuracy'])

        run_props = data['props']

        self.assertLess(run_props['creation_time'], run_props['end_time'])
        self.assertEqual('Run # 1', run_props['name'])
        self.assertEqual('default', run_props['experiment']['name'])
        self.assertEqual(0, len(run_props['tags']))

    def test_run_traces_batch_api(self):
        run = self._find_run_by_name('Run # 1')
        client = self.client
        requested_traces = [
            {'name': 'accuracy', 'context': {'is_training': False}},
            {'name': 'loss', 'context': {'is_training': True, 'subset': 'train'}}
        ]
        response = client.post(f'/api/runs/{run.hash}/metric/get-batch/', json=requested_traces)
        self.assertEqual(200, response.status_code)

        traces_batch = response.json()
        self.assertEqual(2, len(traces_batch))
        self.assertEqual('accuracy', traces_batch[0]['name'])
        self.assertEqual(100, len(traces_batch[0]['values']))
        self.assertEqual(False, traces_batch[0]['context']['is_training'])

        self.assertEqual('loss', traces_batch[1]['name'])
        self.assertEqual(100, len(traces_batch[1]['values']))
        self.assertEqual(True, traces_batch[1]['context']['is_training'])
        self.assertEqual('train', traces_batch[1]['context']['subset'])

    def _find_run_by_name(self, name: str) -> Run:
        repo = self.repo
        for run in repo.iter_runs():
            if run.name == name or run.get('name') == name:
                return run
        return None


class TestRunInfoApi(ApiTestBase):
    @pytest.mark.gh_2267
    def test_run_info_api_with_tags(self):
        """covers https://github.com/aimhubio/aim/issues/2267"""
        run = Run(system_tracking_interval=None)
        run.add_tag('Best Run')
        run.close()

        client = self.client
        response = client.get(f'/api/runs/{run.hash}/info/')
        self.assertEqual(200, response.status_code)

        data = response.json()
        run_props = data['props']

        self.assertEqual(1, len(run_props['tags']))
        self.assertEqual('Best Run', run_props['tags'][0]['name'])
        self.assertIsNone(run_props['tags'][0]['color'])
        self.assertIsNone(run_props['tags'][0]['description'])
        tag_id = run_props['tags'][0]['id']

        # Edit tag
        tag = self.repo.structured_db.find_tag(tag_id)
        tag.description = 'Long description for tag'

        response = client.get(f'/api/runs/{run.hash}/info/')
        self.assertEqual(200, response.status_code)

        data = response.json()
        run_props = data['props']

        self.assertEqual(1, len(run_props['tags']))
        self.assertEqual('Long description for tag', run_props['tags'][0]['description'])




import os
import atexit
import signal
import threading
from typing import Optional

from aim.ext.exception_resistant import exception_resistant
from aim.sdk.legacy.deprecation_warning import deprecated
from aim.ext.resource.configs import DEFAULT_SYSTEM_TRACKING_INT
from aim.ext.resource.tracker import ResourceTracker

from aim.sdk.repo import Repo
from aim.sdk.run import Run


class Session:
    sessions = {}

    _are_exit_listeners_set = False
    _original_sigint_handler = None
    _original_sigterm_handler = None

    @deprecated
    def __init__(self, repo: Optional[str] = None,
                 experiment: Optional[str] = None,
                 flush_frequency: int = 0,  # unused
                 block_termination: bool = True,  # unused
                 run: Optional[str] = None,
                 system_tracking_interval: Optional[int] = DEFAULT_SYSTEM_TRACKING_INT):

        self._repo = Repo.from_path(repo) if repo else Repo.default_repo()
        self._repo_path = self._repo.path
        self._run = Run(run, repo=self._repo, experiment=experiment,
                        system_tracking_interval=system_tracking_interval)
        self._run_hash = self._run.hash
        self.active = True

        Session.sessions.setdefault(self._repo_path, [])
        Session.sessions[self._repo_path].append(self)

        # Bind signal listeners
        self._set_exit_handlers()

    @property
    def run_hash(self):
        return self._run_hash

    @property
    def repo_path(self):
        return self._repo_path

    @exception_resistant(silent=False)
    def track(self, *args, **kwargs):
        val = args[0]
        name = kwargs.pop('name')
        step = kwargs.pop('step', None)
        epoch = kwargs.pop('epoch', None)
        for key in kwargs.keys():
            if key.startswith('__'):
                del kwargs[key]

        self._run.track(val, name=name, step=step, epoch=epoch, context=kwargs)

    @exception_resistant(silent=False)
    def set_params(self, params: dict, name: Optional[str] = None):
        if name is None:
            self._run[...] = params
        else:
            self._run[name] = params

    def flush(self):
        pass

    @exception_resistant(silent=False)
    def close(self):
        if not self.active:
            raise Exception('session is closed')
        if self._run:
            del self._run
            self._run = None
        if self._repo_path in Session.sessions \
                and self in Session.sessions[self._repo_path]:
            Session.sessions[self._repo_path].remove(self)
            if len(Session.sessions[self._repo_path]) == 0:
                del Session.sessions[self._repo_path]
        self.active = False

    @classmethod
    def _close_sessions(cls, *args, **kwargs):
        threads = []
        for _, sessions in cls.sessions.items():
            for session in sessions:
                th = threading.Thread(target=session.close)
                th.daemon = True
                threads.append(th)

        for th in threads:
            th.start()

        for th in threads:
            th.join()

        if len(args):
            if args[0] == 15:
                signal.signal(signal.SIGTERM, cls._original_sigterm_handler)
                os.kill(os.getpid(), 15)
            # elif args[0] == 2:
            #     signal.signal(signal.SIGINT, cls._original_sigint_handler)
            #     os.kill(os.getpid(), 2)

    @classmethod
    def _set_exit_handlers(cls):
        if not cls._are_exit_listeners_set:
            cls._are_exit_listeners_set = True
            # cls._original_sigint_handler = signal.getsignal(signal.SIGINT)
            cls._original_sigterm_handler = signal.getsignal(signal.SIGTERM)

            atexit.register(cls._close_sessions)
            # signal.signal(signal.SIGINT, cls._close_sessions)
            signal.signal(signal.SIGTERM, cls._close_sessions)


DefaultSession = Session


def get_default_session() -> Session:
    if len(Session.sessions.keys()) > 0:
        default_sess_key = list(Session.sessions.keys())[0]
        if len(Session.sessions[default_sess_key]) > 0:
            return Session.sessions[default_sess_key][0]

    # Create and return default session otherwise
    return DefaultSession()



import os
import platform
import shutil
import subprocess
import time
import logging

from typing import Optional, Tuple

logger = logging.getLogger(__name__)


def get_username():
    """
    Utility function to retrieve the current username
    """
    if os.environ.get('LOGNAME'):
        return os.environ.get('LOGNAME')
    if os.environ.get('LNAME'):
        return os.environ.get('LNAME')
    if os.environ.get('USER'):
        return os.environ.get('USER')
    if os.environ.get('USERNAME'):
        return os.environ.get('USERNAME')

    return None


def check_sshfs_installation():
    """
    Utility function to check sshfs installation and user permission to execute sshfs.
    Raises errors with hints to install sshfs based on the platform
    """
    sshfs_executable = shutil.which('sshfs')
    if not sshfs_executable:
        if platform.system().lower() == 'darwin':
            raise subprocess.SubprocessError('Could not find sshfs installation. \n'
                                             'To install sshfs please run the following commands:\n'
                                             'brew install macfuse \n'
                                             'brew install gromgit/fuse/sshfs\n'
                                             'or alternatively follow the instructions here:'
                                             'https://osxfuse.github.io/')
        else:
            raise subprocess.SubprocessError('Could not find sshfs installation. \n'
                                             'To install sshfs please run the following commands:\n'
                                             'apt-get update\n'
                                             'apt-get install sshfs')

    # check user permission to execute sshfs
    if not os.access(sshfs_executable, os.EX_OK):
        username = get_username()
        raise PermissionError(f'User {username} does not have sufficient permissions to run sshfs command.')


def check_directory_permissions(path: str, mount_root: Optional[str] = '', unmount: bool = False):
    """
    Utility function to check user write permissions on given path.
    Optionally can unmount the the given path.
    """
    if not os.access(path, os.W_OK):
        username = get_username()
        if unmount:
            unmount_remote_repo(path, mount_root)
        raise PermissionError(f'User {username} does not have write permissions on path: {path}.')


def extract_remote_user_pwd_host_repo(path: str) -> Tuple[Optional[str], Optional[str], Optional[str], str]:
    """
    Utility function to extract user, password, host and remote repo path from given remote path.
    Arg path should follow the following format: 'ssh://[user:password@][host]:remote_repo_path'
    Falls back to environment variables if some of the extractables are missing
    """
    assert path.startswith('ssh://')
    # strip the 'ssh://' prefix
    path = path[6:]

    # split user:password and host:remote_repo_path pairs
    initial_split = path.split('@', maxsplit=1)

    user = None
    password = None
    host = None

    if len(initial_split) == 1:
        user_pwd_split = None
        host_path_split = initial_split[0].split(':', maxsplit=1)
    else:
        host_path_split = initial_split[1].split(':', maxsplit=1)
        user_pwd_split = initial_split[0].split(':', maxsplit=1)

    # split user and password
    if user_pwd_split:
        user = user_pwd_split[0]
        if len(user_pwd_split) > 1:
            password = user_pwd_split[1]

    # split host and remote path
    if len(host_path_split) == 1:
        remote_path = host_path_split[0]
    else:
        host = host_path_split[0]
        remote_path = host_path_split[1]

    # fallback to env variables
    if not user and os.environ.get('AIM_REMOTE_REPO_USER'):
        user = os.environ.get('AIM_REMOTE_REPO_USER')
    if not password and os.environ.get('AIM_REMOTE_REPO_PASSWORD'):
        password = os.environ.get('AIM_REMOTE_REPO_PASSWORD')
    if not host and os.environ.get('AIM_REMOTE_REPO_HOST'):
        host = os.environ.get('AIM_REMOTE_REPO_HOST')

    return user, password, host, remote_path


def mount_remote_repo(remote_path: str) -> Tuple[str, str]:
    """
    Utility function to mount remote repository path to local  '/tmp/<timestamp>/<remote_repo_name>' directory
    """

    check_sshfs_installation()

    user, pwd, host, remote_repo_path = extract_remote_user_pwd_host_repo(remote_path)

    # check if the current user has permissions to perform write operations on /tmp dir
    check_directory_permissions(path='/tmp')
    mount_root = f'/tmp/{int(time.time())}'
    mount_point = f'{mount_root}/{remote_repo_path}'
    os.makedirs(mount_point, exist_ok=True)

    login_options = f'{host}:{remote_repo_path}'
    if user:
        if pwd:
            login_options = f'{user}:{pwd}@{login_options}'
        else:
            login_options = f'{user}@{login_options}'

    # TODO: experiment on sshfs options to find an optimal set
    # construct sshfs options, use IdentityFile if provided by the user
    sshfs_options = 'default_permissions'
    identity_file = os.environ.get('AIM_REMOTE_REPO_KEY_FILE')
    if identity_file:
        sshfs_options = f'{sshfs_options},IdentityFile={identity_file}'

    # try mounting using sshfs
    cmd = ['sshfs', login_options, mount_point, '-o', sshfs_options]
    sshfs_process = subprocess.Popen(cmd, stderr=subprocess.PIPE)
    # redirect sdterr to pipe and check if errors occurred, if not let the sshfs process run in the background
    # otherwise raise an error, kill the process and cleanup the local mount point
    if sshfs_process.stderr.readline():
        # if mounting fails remove local mount point created by Aim
        # and print the command that was used, so it'll be easier for the user to perform manual mounting
        sshfs_process.wait()
        unmount_remote_repo(mount_point, mount_root)
        raise subprocess.SubprocessError(f'Could not mount remote repository using command: \n'
                                         f'{" ".join(cmd)} \n'
                                         f'Please try to mount manually '
                                         f'and use the local mount point as a usual repo path.')

    # check if the user has permissions to perform write operations on local mount point (which means that the
    # remote user also can perform write operations on remote repo path as well)
    # unmount the local mount point in case of failure
    check_directory_permissions(mount_point, mount_root, unmount=True)

    return mount_root, mount_point


def unmount_remote_repo(mount_point: str, mount_root: str):
    """
    Utility function to unmount remote repo and cleanup local directories created by Aim for mounting
    """

    # TODO: [MV] this is experimental
    # TODO: [MV] decide later what to do with unmounting remote repo
    if not os.path.exists(mount_point):
        return

    if not os.path.ismount(mount_point):
        shutil.rmtree(mount_root)
        return

    # force unmount the remote path as we are sure we've done everything we need up until to this point
    # and don't want to get stopped by any other processes stopping from unmounting
    umount_executable = shutil.which('umount')
    cmd = [umount_executable, mount_point]
    child = subprocess.Popen(
        cmd,
    )
    child.communicate()
    exit_code = child.wait()
    if exit_code != 0:
        # in case of failure log warning so the user can unmount manually if needed
        logger.warning(f'Could not unmount path: {mount_point}.\n'
                       f'Please unmount manually using command:\n'
                       f'{" ".join(cmd)}')
    else:
        shutil.rmtree(mount_root)



# AIM SERVER
AIM_SERVER_DEFAULT_PORT = 53800
AIM_SERVER_DEFAULT_HOST = '0.0.0.0'
AIM_SERVER_MOUNTED_REPO_PATH = '__AIM_SERVER_MOUNT_REPO_PATH__'

# AIM CLIENT
# The path of the PEM-encoded root certificates. Used for secure channel establishment
AIM_CLIENT_SSL_CERTIFICATES_FILE = '__AIM_CLIENT_SSL_CERTIFICATES_FILE__'
AIM_CLIENT_QUEUE_MAX_MEMORY = '__AIM_CLIENT_QUEUE_MAX_MEMORY__'

# GRPC OPTIONS
AIM_RT_MAX_MESSAGE_SIZE = '__AIM_RT_MAX_MESSAGE_SIZE__'
AIM_RT_DEFAULT_MAX_MESSAGE_SIZE = 16 * 1024 * 1024  # 16MB



from tests.base import PrefilledDataTestBase

from aim.sdk.utils import generate_run_hash

class TestStructuredDatabase(PrefilledDataTestBase):
    def test_entity_chaining_syntax(self):
        run = self.repo.structured_db.find_run('missing_run_hash')
        self.assertFalse(run.experiment)
        self.assertFalse(run.experiment.name)
        self.assertFalse(run.tags)
        self.assertFalse(run.name)
        self.assertFalse(run.description)

    def test_entity_relations(self):
        with self.repo.structured_db as db:
            db.create_experiment('my experiment')
            runs = [db.create_run(generate_run_hash()) for _ in range(10)]
            for run in runs:
                run.experiment = 'my experiment'

        for run in runs:
            self.assertEqual('my experiment', run.experiment_obj.name)
            self.assertEqual(10, len(run.experiment_obj.runs))

    def test_context_manager_nesting(self):
        with self.repo.structured_db as db1:
            with self.repo.structured_db as db2:
                db1.create_experiment('exp 1')
                db2.create_experiment('exp 2')

        experiment_names = set((exp.name for exp in self.repo.structured_db.experiments()))
        expected_names = set(('exp 1', 'exp 2'))
        self.assertTrue(experiment_names.issuperset(expected_names))



import click
import os
import tqdm

from multiprocessing.pool import ThreadPool
from psutil import cpu_count

from aim.cli.runs.utils import list_repo_runs, match_runs, make_zip_archive, upload_repo_runs, optimize_container
from aim.sdk.repo import Repo
from aim.sdk.lock_manager import LockManager
from aim.sdk.index_manager import RepoIndexManager


@click.group()
@click.option('--repo', required=False,
              default=os.getcwd(),
              type=click.Path(
                  exists=True,
                  file_okay=False,
                  dir_okay=True,
                  writable=True
              ))
@click.pass_context
def runs(ctx, repo):
    """Manage runs in aim repository."""
    ctx.ensure_object(dict)
    ctx.obj['repo'] = repo


@runs.command(name='ls')
@click.pass_context
def list_runs(ctx):
    """List Runs available in Repo."""
    repo_path = ctx.obj['repo']
    if not Repo.exists(repo_path):
        click.echo(f'\'{repo_path}\' is not a valid aim repo.')
        exit(1)

    run_hashes = list_repo_runs(repo_path)

    click.echo('\t'.join(run_hashes))
    click.echo(f'Total {len(run_hashes)} runs.')


@runs.command(name='rm')
@click.argument('hashes', nargs=-1, type=str)
@click.pass_context
def remove_runs(ctx, hashes):
    """Remove Run data for given run hashes."""
    if len(hashes) == 0:
        click.echo('Please specify at least one Run to delete.')
        exit(1)
    repo_path = ctx.obj['repo']
    repo = Repo.from_path(repo_path)

    matched_hashes = match_runs(repo_path, hashes)
    confirmed = click.confirm(f'This command will permanently delete {len(matched_hashes)} runs from aim repo '
                              f'located at \'{repo_path}\'. Do you want to proceed?')
    if not confirmed:
        return

    success, remaining_runs = repo.delete_runs(matched_hashes)
    if success:
        click.echo(f'Successfully deleted {len(matched_hashes)} runs.')
    else:
        click.echo('Something went wrong while deleting runs. Remaining runs are:', err=True)
        click.secho('\t'.join(remaining_runs), fg='yellow')


@runs.command(name='cp')
@click.option('--destination', required=True,
              type=click.Path(
                  exists=True,
                  file_okay=False,
                  dir_okay=True,
                  writable=True
              ))
@click.argument('hashes', nargs=-1, type=str)
@click.pass_context
def copy_runs(ctx, destination, hashes):
    """Copy Run data for given run hashes to destination Repo."""
    if len(hashes) == 0:
        click.echo('Please specify at least one Run to copy.')
        exit(1)
    source = ctx.obj['repo']
    source_repo = Repo.from_path(source)
    destination_repo = Repo.from_path(destination)

    matched_hashes = match_runs(source, hashes)
    success, remaining_runs = source_repo.copy_runs(matched_hashes, destination_repo)
    if success:
        click.echo(f'Successfully copied {len(matched_hashes)} runs.')
    else:
        click.echo('Something went wrong while copying runs. Remaining runs are:', err=True)
        click.secho('\t'.join(remaining_runs), fg='yellow')


@runs.command(name='mv')
@click.option('--destination', required=True,
              type=click.Path(
                  exists=True,
                  file_okay=False,
                  dir_okay=True,
                  writable=True
              ))
@click.argument('hashes', nargs=-1, type=str)
@click.pass_context
def move_runs(ctx, destination, hashes):
    """Move Run data for given run hashes to destination Repo."""
    if len(hashes) == 0:
        click.echo('Please specify at least one Run to move.')
        exit(1)
    source = ctx.obj['repo']
    source_repo = Repo.from_path(source)
    destination_repo = Repo.from_path(destination)

    matched_hashes = match_runs(source, hashes)

    success, remaining_runs = source_repo.move_runs(matched_hashes, destination_repo)
    if success:
        click.echo(f'Successfully moved {len(matched_hashes)} runs.')
    else:
        click.echo('Something went wrong while moving runs. Remaining runs are:', err=True)
        click.secho('\t'.join(remaining_runs), fg='yellow')


@runs.command(name='upload')
@click.argument('bucket', nargs=1, type=str)
@click.pass_context
def upload_runs(ctx, bucket):
    """Upload Repo backup to the given S3 bucket."""
    repo_path = ctx.obj['repo']
    if not Repo.exists(repo_path):
        click.echo(f'\'{repo_path}\' is not a valid aim repo.')
        exit(1)

    zip_buffer = make_zip_archive(repo_path)
    zip_buffer.seek(0)

    success, uploaded_zip_file_name = upload_repo_runs(zip_buffer, bucket)
    if success:
        click.echo(f'Successfully uploaded runs in {uploaded_zip_file_name}.')
    else:
        click.echo(f'The storage backup failed because of the following error: {uploaded_zip_file_name}.')


@runs.command(name='close')
@click.argument('hashes', nargs=-1, type=str)
@click.pass_context
def close_runs(ctx, hashes):
    """Close failed/stalled Runs."""
    repo_path = ctx.obj['repo']
    repo = Repo.from_path(repo_path)

    if len(hashes) == 0:
        click.echo('Please specify at least one Run to close.')
        exit(1)

    click.secho(f'This command will forcefully close {len(hashes)} Runs from Aim Repo \'{repo_path}\'. '
                f'Please make sure Runs are not active. Data corruption may occur otherwise.')
    if not click.confirm('Do you want to proceed?'):
        return

    lock_manager = LockManager(repo.path)
    index_manager = RepoIndexManager.get_index_manager(repo)

    def close_run(run_hash):
        if lock_manager.release_locks(run_hash, force=True):
            # Run rocksdb optimizations if container locks are removed
            meta_db_path = os.path.join(repo_path, 'meta', 'chunks', run_hash)
            seqs_db_path = os.path.join(repo_path, 'seqs', 'chunks', run_hash)
            optimize_container(meta_db_path, extra_options={'compaction': True})
            optimize_container(seqs_db_path, extra_options={})
        if index_manager.run_needs_indexing(run_hash):
            index_manager.index(run_hash)

    pool = ThreadPool(cpu_count(logical=False))

    for _ in tqdm.tqdm(
            pool.imap_unordered(close_run, hashes),
            desc='Closing runs',
            total=len(hashes)):
        pass



from tensorboard.backend.event_processing.directory_watcher import DirectoryWatcher
from tensorboard.backend.event_processing import event_file_loader
import tensorflow as tf
from tensorboard.util import tensor_util
import time
import threading
from pathlib import Path
import logging
import os
import weakref
import queue

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from aim import Audio, Image


class TensorboardTracker:
    def __init__(self, tracker, sync_tensorboard_log_dir: str) -> None:
        self.tracker = tracker
        self.sync_tensorboard_log_dir = sync_tensorboard_log_dir
        self.tensorboard_folder_watchers = []
        self._thread = threading.Thread(target=self._monitor_eventfiles, daemon=True)
        self.directories_track_status = {}
        self._shutdown = False
        self._started = False
        self._watcher_queue = queue.Queue()

    def _monitor_eventfiles(self):
        while True:
            if self._shutdown:
                break
            for event_file in set(Path(self.sync_tensorboard_log_dir).rglob("*.tfevents*")):
                dir = str(event_file.parent.absolute())
                if dir not in self.directories_track_status:
                    self.directories_track_status[dir] = "NOT_STARTED"
            for dir, status in self.directories_track_status.items():
                if status == "NOT_STARTED":
                    tensorboard_folder_watcher = TensorboardFolderTracker(dir, self._watcher_queue)
                    tensorboard_folder_watcher.start()
                    self.tensorboard_folder_watchers.append(tensorboard_folder_watcher)
                    self.directories_track_status[dir] = "STARTED"
            time.sleep(5)

    def start(self):
        if self._started:
            return
        self._started = True
        self._thread.start()
        self._consumer = TensorboardEventConsumer(
            self._watcher_queue, self.tracker
        )
        self._consumer.start()

    def stop(self):
        if not self._started:
            return
        self._shutdown = True
        self._thread.join()
        for tensorboard_folder_watcher in self.tensorboard_folder_watchers:
            tensorboard_folder_watcher.stop()
        self._consumer.stop()

    def close(self):
        """Interface to make compatible with Resource AutoClean"""
        self.stop()


class TensorboardFolderTracker:
    def __init__(self, tensorboard_event_folder: str, queue: queue.Queue) -> None:
        self.queue = queue
        self.supported_plugins = ("images", "scalars")
        self.unsupported_plugin_noticed = False
        self.folder_name = os.path.basename(tensorboard_event_folder)
        self._thread = threading.Thread(target=self._process_event)
        self._generator = DirectoryWatcher(tensorboard_event_folder, event_file_loader.EventFileLoader)
        self._shutdown = False
        self._started = False

    def start(self):
        if self._started:
            return
        self._started = True
        self._thread.start()

    def stop(self):
        if not self._started:
            return
        self._shutdown = True
        self._thread.join()

    def _process_event(self):
        while True:
            if self._shutdown:
                break
            for event in self._generator.Load():
                self._process_tb_event(event)
            time.sleep(1)

    def _process_tb_event(self, event):
        def create_ndarray(tensor):
            res = tensor_util.make_ndarray(tensor)
            if res.dtype == "object":
                return None
            else:
                return res

        step = event.step
        fail_count = 0
        _err_info = None

        for value in event.summary.value:
            tag = value.tag
            plugin_name = value.metadata.plugin_data.plugin_name
            if len(plugin_name) > 0 and plugin_name not in self.supported_plugins:
                if not self.unsupported_plugin_noticed:
                    logging.warning(
                        "Found unsupported plugin type({}) in the log file. "
                        "Data for these wont be processed. "
                        "Supported plugin types are: {}".format(plugin_name, ", ".join(self.supported_plugins)),
                    )
                    self.unsupported_plugin_noticed = True
                continue
            track_val = None
            try:
                if value.HasField("tensor"):
                    # TODO: [MV] check the case when audios are passed via tensor
                    if plugin_name == "images":
                        tensor = value.tensor.string_val[2:]
                        track_val = [Image(tf.image.decode_image(t).numpy()) for t in tensor]
                        if len(track_val) == 1:
                            track_val = track_val[0]
                    elif plugin_name == "scalars" or plugin_name == "":
                        track_val = create_ndarray(value.tensor)
                    else:
                        track_val = value.tensor.float_val[0]
                elif value.HasField("simple_value"):
                    track_val = value.simple_value
                elif value.HasField("image"):
                    track_val = Image(tf.image.decode_image(value.image.encoded_image_string).numpy())
                elif value.HasField("audio"):
                    tf_audio, sample_rate = tf.audio.decode_wav(value.audio.encoded_audio_string)
                    track_val = Audio(tf_audio.numpy(), rate=sample_rate)

            except RuntimeError as exc:
                # catch all the nasty failures
                fail_count += 1
                if not _err_info:
                    _err_info = str(exc)
                continue

            if track_val is not None:
                self.queue.put(TensorboardEvent(track_val, tag, step, context={'entry': self.folder_name}))
        if fail_count:
            logging.warning(f"Failed to process {fail_count} entries. First exception: {_err_info}")


class TensorboardEvent:

    def __init__(self, value: Any, name: str, step: int, context: dict) -> None:
        self.value = value
        self.name = name
        self.step = step
        self.context = context


class TensorboardEventConsumer:

    def __init__(self, queue: queue.Queue, tracker) -> None:
        self._tracker = weakref.ref(tracker)
        self._queue = queue
        self._thread = threading.Thread(target=self._process_events, daemon=True)
        self._shutdown = False
        self._started = False

    def start(self):
        if self._started:
            return
        self._started = True
        self._thread.start()

    def _process_events(self):
        while True:
            try:
                event = self._queue.get(True, 1)
                if event:
                    self._tracker()(event.value, event.name, event.step, context=event.context)
            except queue.Empty:
                event = None
                if self._shutdown:
                    break

    def stop(self):
        if not self._started:
            return
        self._shutdown = True
        self._thread.join()



# distutils: language = c++
# cython: language_level = 3

# We define a mapping for primitive types into a single bit
# The types are going to be stored in a single byte, so 7 more bits are reserved
cdef enum FLAGS:
    _NONE = 0
    _BOOL = 1
    _INT = 2
    _FLOAT = 3
    _STRING = 4
    _BYTES = 5
    _ARRAY = 6
    _OBJECT = 7
    _CUSTOM_OBJECT = 8 | 7


cpdef object encode(object value)

cpdef object decode(object buffer)

cpdef bytes encode_key(object key)

cpdef bytes encode_path(object path)



from typing import Optional, Union

from aim.sdk.run import BasicRun
from aim.ext.tensorboard_tracker.tracker import TensorboardTracker

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from aim.sdk.repo import Repo


class Run(BasicRun):
    def __init__(self, run_hash: Optional[str] = None, *,
                 sync_tensorboard_log_dir: str,
                 repo: Optional[Union[str, 'Repo']] = None,
                 experiment: Optional[str] = None,
                 force_resume: Optional[bool] = False,
                 ):
        super().__init__(run_hash, repo=repo, read_only=False, experiment=experiment, force_resume=force_resume)
        self['tb_log_directory'] = sync_tensorboard_log_dir
        self._tensorboard_tracker = TensorboardTracker(self._tracker, sync_tensorboard_log_dir)
        self._tensorboard_tracker.start()
        self._resources.add_extra_resource(self._tensorboard_tracker)



# distutils: language = c++
# cython: language_level = 3
cimport cython

from aim.storage.container cimport Container
from aim.storage.container cimport ContainerItemsIterator
from aim.storage.containertreeview cimport ContainerTreeView

cdef class PrefixView(Container):
    cdef public bytes prefix
    # TODO cdef public Container parent after all container bindings are ready
    cdef public parent
    cdef public bint read_only

    cpdef bytes absolute_path(self, bytes path = *)
    cpdef ContainerTreeView tree(self)

cdef class PrefixViewItemsIterator(ContainerItemsIterator):
    cdef PrefixView prefix_view
    cdef bytes path
    cdef int prefix_len
    cdef ContainerItemsIterator it

    @cython.locals(item=tuple, keys=bytes)
    cpdef object next(self)



# Aim
AIM_OBJECTS_DIR_NAME = 'objects'
AIM_METRICS_DIR_NAME = 'metrics'
AIM_METRIC_GR_DIR_NAME = 'metric_groups'
AIM_ANNOT_DIR_NAME = 'annotations'
AIM_SEG_DIR_NAME = 'segmentation'
AIM_MEDIA_DIR_NAME = 'media'
AIM_IMAGES_DIR_NAME = 'images'
AIM_MODELS_DIR_NAME = 'models'
AIM_CORR_DIR_NAME = 'correlations'
AIM_CORR_DIRS_NAME = 'dirs'
AIM_CONFIG_FILE_NAME = 'config.json'
AIM_DIFF_DIR_NAME = 'diff'
AIM_DIFF_FILE_NAME = 'diff.txt'
AIM_PARAMS_DIR_NAME = 'hyperparameters'
AIM_MAP_DIR_NAME = 'map'
AIM_STATS_DIR_NAME = 'stats'
AIM_TEXT_DIR_NAME = 'document_texts'
AIM_MAP_METRICS_KEYWORD = '__METRICS__'

# Commits
AIM_COMMIT_INDEX_DIR_NAME = 'index'
AIM_COMMIT_CONFIG_FILE_NAME = 'config.json'
AIM_COMMIT_META_FILE_NAME = 'meta.json'



from typing import Optional

from aim.sdk.repo import Repo
from aim.sdk.legacy.deprecation_warning import deprecated


@deprecated
def select_metrics(search_statement: str, repo_path: Optional[str] = None):
    if repo_path is not None:
        repo = Repo.from_path(repo_path)
    else:
        repo = Repo.default_repo()

    if not repo:
        return None

    return repo.query_metrics(search_statement)


@deprecated
def select_runs(expression: Optional[str] = None, repo_path: Optional[str] = None):
    if repo_path is not None:
        repo = Repo.from_path(repo_path)
    else:
        repo = Repo.default_repo()

    if not repo:
        return None

    return repo.query_runs(expression)



from typing import Dict, Any

from aim.sdk.run import Run
from aim.sdk.callbacks import Caller, event


# TODO [AT]: This is a draft version. Final list of methods/signatures TBD.
class TrainingFlow(Caller):
    """High-level Run object, resembling the execution flow of the typical training script."""
    def __init__(self, run: Run, callbacks=None):
        super().__init__(callbacks=callbacks)
        self.run = run
        self.init()

    def _extra_kwargs(self) -> Dict[str, Any]:
        kwargs = super()._extra_kwargs()
        kwargs['run'] = self.run

        return kwargs

    @event
    def init(
        self,
        **kwargs
    ):
        """Guaranteed to be called before any other event."""

    @event
    def training_started(
        self, *,
        hparams: Dict[str, Any],
        run: Run,
        **kwargs
    ):
        """Is called just after Run object and hyperparameters are ready."""

    @event
    def training_metrics_collected(
        self, *,
        metrics: Dict[str, Any], step: int, epoch: int = None,
        run: Run,
        **kwargs
    ):
        """Is called after the training metrics are calculated and ready to be logged."""

    @event
    def validation_metrics_collected(
        self, *,
        metrics: Dict[str, Any], step: int, epoch: int = None,
        run: Run,
        **kwargs
    ):
        """Is called after the validation metrics are calculated and ready to be logged."""

    @event
    def training_successfully_finished(
        self, *,
        run: Run,
        **kwargs
    ):
        """Is called after the training phase is successfully finished."""

    # @event
    # def run_available_experimental(
    #     self, *,
    #     run: Run,
    #     **kwargs
    # ):
    #     ...



"""empty message

Revision ID: 73a3d004c227
Revises: 5ae8371b7481
Create Date: 2021-08-31 20:57:56.503417

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '73a3d004c227'
down_revision = '5ae8371b7481'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('commits')
    op.drop_table('explore_states')
    op.drop_table('commit_tag')
    op.drop_table('tf_summary_logs')
    op.drop_table('dashboards')
    op.drop_table('tags')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('tags',
    sa.Column('uuid', sa.TEXT(), nullable=False),
    sa.Column('name', sa.TEXT(), nullable=True),
    sa.Column('color', sa.TEXT(), nullable=True),
    sa.Column('created_at', sa.DATETIME(), nullable=True),
    sa.Column('is_archived', sa.BOOLEAN(), nullable=True),
    sa.Column('is_hidden', sa.BOOLEAN(), nullable=True),
    sa.PrimaryKeyConstraint('uuid')
    )
    op.create_table('dashboards',
    sa.Column('uuid', sa.TEXT(), nullable=False),
    sa.Column('name', sa.TEXT(), nullable=True),
    sa.Column('description', sa.TEXT(), nullable=True),
    sa.Column('created_at', sa.DATETIME(), nullable=True),
    sa.Column('updated_at', sa.DATETIME(), nullable=True),
    sa.Column('is_archived', sa.BOOLEAN(), nullable=True),
    sa.PrimaryKeyConstraint('uuid')
    )
    op.create_table('tf_summary_logs',
    sa.Column('uuid', sa.TEXT(), nullable=False),
    sa.Column('log_path', sa.TEXT(), nullable=True),
    sa.Column('params', sa.TEXT(), nullable=True),
    sa.Column('created_at', sa.DATETIME(), nullable=True),
    sa.Column('is_archived', sa.BOOLEAN(), nullable=True),
    sa.PrimaryKeyConstraint('uuid')
    )
    op.create_table('commit_tag',
    sa.Column('commit_id', sa.TEXT(), nullable=True),
    sa.Column('tag_id', sa.TEXT(), nullable=True),
    sa.ForeignKeyConstraint(['commit_id'], ['commits.uuid'], ),
    sa.ForeignKeyConstraint(['tag_id'], ['tags.uuid'], )
    )
    op.create_table('explore_states',
    sa.Column('uuid', sa.TEXT(), nullable=False),
    sa.Column('created_at', sa.DATETIME(), nullable=True),
    sa.Column('updated_at', sa.DATETIME(), nullable=True),
    sa.Column('is_archived', sa.BOOLEAN(), nullable=True),
    sa.Column('chart_focused_step', sa.INTEGER(), nullable=True),
    sa.Column('chart_focused_metric_run_hash', sa.TEXT(), nullable=True),
    sa.Column('chart_focused_metric_metric_name', sa.TEXT(), nullable=True),
    sa.Column('chart_focused_metric_trace_context', sa.TEXT(), nullable=True),
    sa.Column('chart_focused_circle_active', sa.BOOLEAN(), nullable=True),
    sa.Column('chart_focused_circle_run_hash', sa.TEXT(), nullable=True),
    sa.Column('chart_focused_circle_metric_name', sa.TEXT(), nullable=True),
    sa.Column('chart_focused_circle_step', sa.INTEGER(), nullable=True),
    sa.Column('chart_focused_circle_trace_context', sa.TEXT(), nullable=True),
    sa.Column('chart_focused_circle_param', sa.TEXT(), nullable=True),
    sa.Column('chart_focused_circle_content_type', sa.TEXT(), nullable=True),
    sa.Column('chart_settings_zoom_mode', sa.BOOLEAN(), nullable=True),
    sa.Column('chart_settings_single_zoom_mode', sa.BOOLEAN(), nullable=True),
    sa.Column('chart_settings_zoom_history', sa.TEXT(), nullable=True),
    sa.Column('chart_settings_highlight_mode', sa.TEXT(), nullable=True),
    sa.Column('chart_settings_persistent_display_outliers', sa.BOOLEAN(), nullable=True),
    sa.Column('chart_settings_persistent_zoom', sa.TEXT(), nullable=True),
    sa.Column('chart_settings_persistent_interpolate', sa.BOOLEAN(), nullable=True),
    sa.Column('chart_settings_persistent_indicator', sa.BOOLEAN(), nullable=True),
    sa.Column('chart_settings_persistent_x_alignment', sa.TEXT(), nullable=True),
    sa.Column('chart_settings_persistent_x_scale', sa.INTEGER(), nullable=True),
    sa.Column('chart_settings_persistent_y_scale', sa.INTEGER(), nullable=True),
    sa.Column('chart_settings_persistent_points_count', sa.INTEGER(), nullable=True),
    sa.Column('chart_settings_persistent_smoothing_algorithm', sa.TEXT(), nullable=True),
    sa.Column('chart_settings_persistent_smooth_factor', sa.FLOAT(), nullable=True),
    sa.Column('chart_settings_persistent_aggregated', sa.BOOLEAN(), nullable=True),
    sa.Column('chart_hidden_metrics', sa.TEXT(), nullable=True),
    sa.Column('chart_tooltip_options_display', sa.BOOLEAN(), nullable=True),
    sa.Column('chart_tooltip_options_fields', sa.TEXT(), nullable=True),
    sa.Column('search_query', sa.TEXT(), nullable=True),
    sa.Column('search_v', sa.INTEGER(), nullable=True),
    sa.Column('search_input_value', sa.TEXT(), nullable=True),
    sa.Column('search_input_select_input', sa.TEXT(), nullable=True),
    sa.Column('search_input_select_condition_input', sa.TEXT(), nullable=True),
    sa.Column('context_filter_group_by_color', sa.TEXT(), nullable=True),
    sa.Column('context_filter_group_by_style', sa.TEXT(), nullable=True),
    sa.Column('context_filter_group_by_chart', sa.TEXT(), nullable=True),
    sa.Column('context_filter_group_against_color', sa.BOOLEAN(), nullable=True),
    sa.Column('context_filter_group_against_style', sa.BOOLEAN(), nullable=True),
    sa.Column('context_filter_group_against_chart', sa.BOOLEAN(), nullable=True),
    sa.Column('context_filter_aggregated_area', sa.TEXT(), nullable=True),
    sa.Column('context_filter_aggregated_line', sa.TEXT(), nullable=True),
    sa.Column('context_filter_seed_color', sa.INTEGER(), nullable=True),
    sa.Column('context_filter_seed_style', sa.INTEGER(), nullable=True),
    sa.Column('context_filter_persist_color', sa.BOOLEAN(), nullable=True),
    sa.Column('context_filter_persist_style', sa.BOOLEAN(), nullable=True),
    sa.Column('color_palette', sa.INTEGER(), nullable=True),
    sa.Column('sort_fields', sa.TEXT(), nullable=True),
    sa.Column('table_row_height_mode', sa.TEXT(), nullable=True),
    sa.Column('table_columns_order_left', sa.TEXT(), nullable=True),
    sa.Column('table_columns_order_middle', sa.TEXT(), nullable=True),
    sa.Column('table_columns_order_right', sa.TEXT(), nullable=True),
    sa.Column('table_columns_widths', sa.TEXT(), nullable=True),
    sa.Column('table_excluded_fields', sa.TEXT(), nullable=True),
    sa.Column('screen_view_mode', sa.TEXT(), nullable=True),
    sa.Column('screen_panel_flex', sa.FLOAT(), nullable=True),
    sa.Column('dashboard_id', sa.TEXT(), nullable=True),
    sa.ForeignKeyConstraint(['dashboard_id'], ['dashboards.uuid'], ),
    sa.PrimaryKeyConstraint('uuid')
    )
    op.create_table('commits',
    sa.Column('uuid', sa.TEXT(), nullable=False),
    sa.Column('hash', sa.TEXT(), nullable=True),
    sa.Column('experiment_name', sa.TEXT(), nullable=True),
    sa.Column('session_started_at', sa.INTEGER(), nullable=True),
    sa.Column('session_closed_at', sa.INTEGER(), nullable=True),
    sa.Column('created_at', sa.DATETIME(), nullable=True),
    sa.Column('is_archived', sa.BOOLEAN(), nullable=True),
    sa.PrimaryKeyConstraint('uuid'),
    sa.UniqueConstraint('experiment_name', 'hash')
    )
    # ### end Alembic commands ###



import logging
from typing import Optional

from aim.ext.notifier.base_notifier import BaseNotifier


class LoggingNotifier(BaseNotifier):
    def __init__(self, _id: str, config: dict):
        super().__init__(_id)
        self.message_template = config['message']
        self.logger = logging.getLogger('notifier')

    def notify(self, message: Optional[str] = None, **kwargs):
        message_template = message or self.message_template
        msg = message_template.format(**kwargs)
        self.logger.error(msg)



import os
from datetime import timedelta
from typing import Optional, Tuple

from collections import Counter
from fastapi import Depends, HTTPException, Query, Header
from aim.web.api.utils import APIRouter  # wrapper for fastapi.APIRouter

from aim.web.configs import AIM_UI_TELEMETRY_KEY, AIM_PROJECT_SETTINGS_FILE
from aim.web.api.projects.project import Project
from aim.web.api.projects.pydantic_models import (
    ProjectActivityApiOut,
    ProjectApiOut,
    ProjectParamsOut,
    ProjectPinnedSequencesApiIn,
    ProjectPinnedSequencesApiOut,
)
from aim.web.api.utils import object_factory
from aim.sdk.index_manager import RepoIndexManager
from aim.storage.locking import AutoFileLock

projects_router = APIRouter()


@projects_router.get('/', response_model=ProjectApiOut)
async def project_api():
    project = Project()

    if not project.exists():
        raise HTTPException(status_code=404)

    return {
        'name': project.name,
        'path': project.path,
        'description': project.description,
        'telemetry_enabled': os.getenv(AIM_UI_TELEMETRY_KEY, '1'),
    }


@projects_router.get('/activity/', response_model=ProjectActivityApiOut)
async def project_activity_api(x_timezone_offset: int = Header(default=0),
                               factory=Depends(object_factory)):
    project = Project()

    if not project.exists():
        raise HTTPException(status_code=404)

    num_runs = 0
    num_archived_runs = 0
    activity_counter = Counter()
    for run in factory.runs():
        creation_time = run.created_at - timedelta(minutes=x_timezone_offset)
        activity_counter[creation_time.strftime('%Y-%m-%dT%H:00:00')] += 1
        num_runs += 1
        if run.archived:
            num_archived_runs += 1

    return {
        'num_experiments': len(factory.experiments()),
        'num_runs': num_runs,
        'num_archived_runs': num_archived_runs,
        'num_active_runs': len(project.repo.list_active_runs()),
        'activity_map': dict(activity_counter),
    }


@projects_router.get('/pinned-sequences/', response_model=ProjectPinnedSequencesApiOut)
async def get_pinned_metrics_api():
    import json
    project = Project()

    if not project.exists():
        raise HTTPException(status_code=404)

    response = {'sequences': []}
    settings_filename = os.path.join(project.repo_path, AIM_PROJECT_SETTINGS_FILE)
    settings_lock_file = f'{settings_filename}.lock'
    if not os.path.exists(settings_filename):
        return response

    with AutoFileLock(settings_lock_file, timeout=2):
        with open(settings_filename, 'r') as settings_file:
            try:
                settings = json.load(settings_file)
            except json.decoder.JSONDecodeError:
                return response
            sequences_list = settings.get('pinned_sequences', {}).get('metric', [])
            response['sequences'] = sequences_list
            return response


@projects_router.post('/pinned-sequences/', response_model=ProjectPinnedSequencesApiOut)
async def update_pinned_metrics_api(request_data: ProjectPinnedSequencesApiIn):
    import json
    project = Project()

    if not project.exists():
        raise HTTPException(status_code=404)

    request_dict = request_data.dict()
    sequences_list = request_dict.get('sequences', [])

    # read current settings
    settings_filename = os.path.join(project.repo_path, AIM_PROJECT_SETTINGS_FILE)
    settings_lock_file = f'{settings_filename}.lock'
    settings = {}
    with AutoFileLock(settings_lock_file, timeout=2):
        mode = 'r+'
        if not os.path.exists(settings_filename):
            mode = 'w+'
        with open(settings_filename, mode) as settings_file:
            try:
                settings = json.load(settings_file)
            except json.decoder.JSONDecodeError:
                pass
            settings['pinned_sequences'] = {'metric': sequences_list}
            settings_file.seek(0)
            # dump new settings
            json.dump(settings, settings_file)
            settings_file.truncate()

    return {'sequences': sequences_list}


@projects_router.get('/params/', response_model=ProjectParamsOut, response_model_exclude_defaults=True)
async def project_params_api(sequence: Optional[Tuple[str, ...]] = Query(()),
                             exclude_params: Optional[bool] = False):
    project = Project()

    if not project.exists():
        raise HTTPException(status_code=404)

    if sequence != ():
        try:
            project.repo.validate_sequence_types(sequence)
        except ValueError as e:
            raise HTTPException(status_code=400, detail=str(e))
    else:
        sequence = project.repo.available_sequence_types()
    if exclude_params:
        response = {}
    else:
        response = {
            'params': project.repo.collect_params_info(),
        }
    response.update(**project.repo.collect_sequence_info(sequence))
    return response


@projects_router.get('/status/')
async def project_status_api():
    project = Project()

    if not project.exists():
        raise HTTPException(status_code=404)

    return RepoIndexManager.get_index_manager(project.repo).repo_status



import os

import click
from click import ClickException

from aim.sdk.repo import Repo
from aim.sdk.utils import clean_repo_path
from aim.cli.convert.processors import (
    parse_tb_logs,
    parse_mlflow_logs,
    parse_wandb_logs,
)


@click.group()
@click.option('--repo', required=False, type=str)
@click.pass_context
def convert(ctx, repo):
    ctx.ensure_object(dict)
    if repo and (not repo.startswith('aim:')):
        repo_path = clean_repo_path(repo) or Repo.default_repo_path()
        repo_inst = Repo.from_path(repo_path)
    else:
        repo_inst = repo
    ctx.obj['repo_inst'] = repo_inst


@convert.command(name='tensorboard')
@click.pass_context
@click.option('--logdir', required=True, type=click.Path(exists=True,
                                                         readable=True,
                                                         dir_okay=True,
                                                         resolve_path=True))
@click.option('--flat', '-f', required=False, is_flag=True, default=False)
@click.option('--no-cache', required=False, is_flag=True, default=False)
def convert_tensorboard(ctx, logdir, flat, no_cache):
    repo_inst = ctx.obj['repo_inst']
    parse_tb_logs(logdir, repo_inst, flat, no_cache)


@convert.command(name='tf')
@click.pass_context
@click.option('--logdir', required=True, type=click.Path(exists=True,
                                                         readable=True,
                                                         dir_okay=True,
                                                         resolve_path=True))
@click.option('--flat', '-f', required=False, is_flag=True, default=False)
def convert_tensorflow(ctx, logdir, flat):
    click.secho('WARN: Command \'tf\' is deprecated and will be removed in future releases,'
                ' please use \'tensorboard\' instead.', fg='red')
    repo_inst = ctx.obj['repo_inst']
    parse_tb_logs(logdir, repo_inst, flat)


@convert.command(name='mlflow')
@click.pass_context
@click.option('--tracking_uri', required=False, default=None)
@click.option('--experiment', '-e', required=False, default=None)
def convert_mlflow(ctx, tracking_uri=None, **kwargs):
    repo_inst = ctx.obj['repo_inst']
    tracking_uri = tracking_uri or os.environ.get("MLFLOW_TRACKING_URI")
    if not tracking_uri:
        raise ClickException("MLFlow tracking URI must be provided either trough ENV or CLI.")
    parse_mlflow_logs(repo_inst, tracking_uri, **kwargs)


@convert.command(name='wandb')
@click.pass_context
@click.option('--entity', required=True, default=None)
@click.option('--project', required=True, default=None)
@click.option('--run-id', required=False, default=None)
def convert_wandb(ctx, entity=None, project=None, **kwargs):
    repo_inst = ctx.obj['repo_inst']
    parse_wandb_logs(repo_inst, entity, project, **kwargs)



import boto3
import os
import shutil
import tarfile
import time
from pathlib import Path

from aim.sdk.configs import AIM_REPO_NAME
from performance_tests.utils import get_baseline_filename

TEST_REPO_PATHS = {
    'real_life_repo': '.aim_performance_repo_1',
    'generated_repo': '.aim_performance_repo_2'
}
AIM_PERFORMANCE_BUCKET_NAME = 'aim-demo-logs'
AIM_PERFORMANCE_LOG_FILE_NAME = 'performance-logs.tar.gz'


def _init_test_repos():
    for repo_path in TEST_REPO_PATHS.values():
        if os.path.exists(repo_path):
            _cleanup_test_repo(repo_path)

    tarfile_name = f'data/{AIM_PERFORMANCE_LOG_FILE_NAME}'
    # download the archive
    if not os.path.exists(tarfile_name):
        Path('data').mkdir(exist_ok=True)
        s3 = boto3.client('s3')
        # needs `AWS_ACCESS_KEY_ID` and `AWS_SECRET_ACCESS_KEY` env variables set up to run locally
        s3.download_file(AIM_PERFORMANCE_BUCKET_NAME,
                         AIM_PERFORMANCE_LOG_FILE_NAME,
                         tarfile_name)
    # extract the archive
    tar = tarfile.open(tarfile_name, 'r:gz')
    tar.extractall()
    tar.close()


def _cleanup_test_repo(path):
    shutil.rmtree(path)


def pytest_sessionstart(session):
    if os.environ.get('AIM_LOCAL_PERFORMANCE_TEST'):
        _init_test_repos()
    else:
        # github actions performance tests on self hosted runner
        os.chdir('/home/ubuntu/performance_logs/')
    time.sleep(10)

def print_current_baseline():
    print('==== CURRENT BASELINE ====')
    with open(get_baseline_filename(), 'r') as f:
        print(f.read())
    print('==========================')

def pytest_unconfigure(config):
    print_current_baseline()

def pytest_sessionfinish(session, exitstatus):
    if os.environ.get('AIM_LOCAL_PERFORMANCE_TEST'):
        for path in TEST_REPO_PATHS.values():
            _cleanup_test_repo(path)
    if os.environ.get(AIM_REPO_NAME):
        del os.environ[AIM_REPO_NAME]



"""soft delete

Revision ID: fbfe5c4702fb
Revises: 1ecf8222220d
Create Date: 2021-08-17 21:34:41.024743

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'fbfe5c4702fb'
down_revision = '1ecf8222220d'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('experiment', sa.Column('is_archived', sa.Boolean(), nullable=True))
    op.add_column('tag', sa.Column('is_archived', sa.Boolean(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('tag', 'is_archived')
    op.drop_column('experiment', 'is_archived')
    # ### end Alembic commands ###



from fastapi import Depends, HTTPException, Query, Header
from fastapi.responses import JSONResponse, StreamingResponse
from starlette import status

from aim.web.api.runs.object_views import (
    ImageApiConfig,
    TextApiConfig,
    DistributionApiConfig,
    AudioApiConfig,
    FigureApiConfig
)
from aim.web.api.utils import APIRouter  # wrapper for fastapi.APIRouter
from typing import Optional, Tuple

from aim.sdk.types import QueryReportMode
from aim.web.api.runs.utils import (
    checked_query,
    collect_requested_metric_traces,
    custom_aligned_metrics_streamer,
    get_project_repo,
    get_run_or_404,
    get_run_params,
    get_run_props,
    metric_search_result_streamer,
    run_active_result_streamer,
    run_search_result_streamer,
    run_logs_streamer

)
from aim.web.api.runs.pydantic_models import (
    MetricAlignApiIn,
    QuerySyntaxErrorOut,
    RunActiveOut,
    RunTracesBatchApiIn,
    RunMetricCustomAlignApiOut,
    RunMetricSearchApiOut,
    RunInfoOut,
    RunsBatchIn,
    RunSearchApiOut,
    RunMetricsBatchApiOut,
    StructuredRunUpdateIn,
    StructuredRunUpdateOut,
    StructuredRunAddTagIn,
    StructuredRunAddTagOut,
    StructuredRunRemoveTagOut,
    StructuredRunsArchivedOut,
    NoteIn,
)
from aim.web.api.utils import object_factory

runs_router = APIRouter()

# static msgs
NOTE_NOT_FOUND = 'Note with id {id} is not found in this run.'


@runs_router.get('/search/run/', response_model=RunSearchApiOut,
                 responses={400: {'model': QuerySyntaxErrorOut}})
async def run_search_api(q: Optional[str] = '',
                         limit: Optional[int] = 0,
                         offset: Optional[str] = None,
                         skip_system: Optional[bool] = True,
                         report_progress: Optional[bool] = True,
                         exclude_params: Optional[bool] = False,
                         exclude_traces: Optional[bool] = False,
                         x_timezone_offset: int = Header(default=0),):
    from aim.sdk.sequence_collection import QueryRunSequenceCollection
    repo = get_project_repo()
    query = checked_query(q)

    repo._prepare_runs_cache()
    runs = QueryRunSequenceCollection(repo=repo,
                                      query=query,
                                      paginated=bool(limit),
                                      offset=offset,
                                      report_mode=QueryReportMode.PROGRESS_TUPLE,
                                      timezone_offset=x_timezone_offset)

    streamer = run_search_result_streamer(runs, limit,
                                          skip_system, report_progress,
                                          exclude_params, exclude_traces)
    return StreamingResponse(streamer)


@runs_router.post('/search/metric/align/', response_model=RunMetricCustomAlignApiOut)
async def run_metric_custom_align_api(request_data: MetricAlignApiIn):
    repo = get_project_repo()
    x_axis_metric_name = request_data.align_by
    requested_runs = request_data.runs

    streamer = custom_aligned_metrics_streamer(requested_runs, x_axis_metric_name, repo)
    return StreamingResponse(streamer)


@runs_router.get('/search/metric/', response_model=RunMetricSearchApiOut,
                 responses={400: {'model': QuerySyntaxErrorOut}})
async def run_metric_search_api(q: Optional[str] = '',
                                p: Optional[int] = 50,
                                x_axis: Optional[str] = None,
                                skip_system: Optional[bool] = True,
                                report_progress: Optional[bool] = True,
                                x_timezone_offset: int = Header(default=0),):
    from aim.sdk.sequences.metric import Metric
    from aim.sdk.sequence_collection import QuerySequenceCollection

    steps_num = p

    if x_axis:
        x_axis = x_axis.strip()

    repo = get_project_repo()
    query = checked_query(q)

    repo._prepare_runs_cache()
    traces = QuerySequenceCollection(repo=repo,
                                     seq_cls=Metric,
                                     query=query,
                                     report_mode=QueryReportMode.PROGRESS_TUPLE,
                                     timezone_offset=x_timezone_offset,)

    streamer = metric_search_result_streamer(traces, skip_system, steps_num, x_axis, report_progress)
    return StreamingResponse(streamer)


@runs_router.get('/active/', response_model=RunActiveOut)
async def get_active_runs_api(report_progress: Optional[bool] = True):
    repo = get_project_repo()
    repo._prepare_runs_cache()

    streamer = run_active_result_streamer(repo)

    return StreamingResponse(streamer)


@runs_router.get('/{run_id}/info/', response_model=RunInfoOut)
async def run_params_api(run_id: str,
                         skip_system: Optional[bool] = False,
                         sequence: Optional[Tuple[str, ...]] = Query(())):
    repo = get_project_repo()
    run = get_run_or_404(run_id, repo=repo)

    if sequence != ():
        try:
            repo.validate_sequence_types(sequence)
        except ValueError as e:
            raise HTTPException(status_code=400, detail=str(e))
    else:
        sequence = repo.available_sequence_types()

    response = {
        'params': get_run_params(run, skip_system=skip_system),
        'traces': run.collect_sequence_info(sequence, skip_last_value=True),
        'props': get_run_props(run)
    }

    response['props'].update({
        'notes': len(run.props.notes_obj)
    })
    return response


@runs_router.post('/{run_id}/metric/get-batch/', response_model=RunMetricsBatchApiOut)
async def run_metric_batch_api(run_id: str, requested_traces: RunTracesBatchApiIn):
    run = get_run_or_404(run_id)
    traces_data = collect_requested_metric_traces(run, requested_traces)

    return JSONResponse(traces_data)


@runs_router.put('/{run_id}/', response_model=StructuredRunUpdateOut)
async def update_run_properties_api(run_id: str, run_in: StructuredRunUpdateIn, factory=Depends(object_factory)):
    with factory:
        run = factory.find_run(run_id)
        if not run:
            raise HTTPException(status_code=404)

        if run_in.name is not None:
            run.name = run_in.name.strip()
        if run_in.description is not None:
            run.description = run_in.description.strip()
        if run_in.experiment is not None:
            run.experiment = run_in.experiment.strip()
        run.archived = run_in.archived

    return {
        'id': run.hash,
        'status': 'OK'
    }


@runs_router.post('/{run_id}/tags/new/', response_model=StructuredRunAddTagOut)
async def add_run_tag_api(run_id: str, tag_in: StructuredRunAddTagIn, factory=Depends(object_factory)):
    with factory:
        run = factory.find_run(run_id)
        if not run:
            raise HTTPException(status_code=404)

        run.add_tag(tag_in.tag_name)
        tag = next(iter(factory.search_tags(tag_in.tag_name)))
    return {
        'id': run.hash,
        'tag_id': tag.uuid,
        'status': 'OK'
    }


@runs_router.delete('/{run_id}/tags/{tag_id}/', response_model=StructuredRunRemoveTagOut)
async def remove_run_tag_api(run_id: str, tag_id: str, factory=Depends(object_factory)):
    with factory:
        run = factory.find_run(run_id)
        tag = factory.find_tag(tag_id)
        if not (run or tag):
            raise HTTPException(status_code=404)

        removed = run.remove_tag(tag.name)

    return {
        'id': run.hash,
        'removed': removed,
        'status': 'OK'
    }


@runs_router.delete('/{run_id}/')
async def delete_run_api(run_id: str):
    repo = get_project_repo()
    success = repo.delete_run(run_id)
    if not success:
        raise HTTPException(status_code=400, detail={
            'message': 'Error while deleting run.',
            'detail': {
                'Run id': run_id
            }
        })

    return {
        'id': run_id,
        'status': 'OK'
    }


@runs_router.post('/delete-batch/')
async def delete_runs_batch_api(runs_batch: RunsBatchIn):
    repo = get_project_repo()
    success, remaining_runs = repo.delete_runs(runs_batch)
    if not success:
        raise HTTPException(status_code=400, detail={
            'message': 'Error while deleting runs.',
            'detail': {
                'Remaining runs id': remaining_runs
            }
        })

    return {
        'status': 'OK'
    }


@runs_router.post('/archive-batch/', response_model=StructuredRunsArchivedOut)
async def archive_runs_batch_api(runs_batch: RunsBatchIn, archive: Optional[bool] = True,
                                 factory=Depends(object_factory)):
    with factory:
        runs = factory.find_runs(runs_batch)
        if not runs:
            raise HTTPException(status_code=404)

        for run in runs:
            run.archived = archive

    return {
        'status': 'OK'
    }


# Note APIs

@runs_router.get('/{run_id}/note/')
def list_note_api(run_id, factory=Depends(object_factory)):
    with factory:
        run = factory.find_run(run_id)
        if not run:
            raise HTTPException(status_code=404)

        notes = run.notes

    return notes


@runs_router.post('/{run_id}/note/', status_code=status.HTTP_201_CREATED)
def create_note_api(run_id, note_in: NoteIn, factory=Depends(object_factory)):
    with factory:
        run = factory.find_run(run_id)
        if not run:
            raise HTTPException(status_code=404)

        note_content = note_in.content.strip()
        note = run.add_note(note_content)

    return {
        'id': note.id,
        'created_at': note.created_at,
    }


@runs_router.get('/{run_id}/note/{_id}')
def get_note_api(run_id, _id: int, factory=Depends(object_factory)):
    with factory:
        run = factory.find_run(run_id)
        if not run:
            raise HTTPException(status_code=404)

        note = run.find_note(_id=_id)
        if not note:
            raise HTTPException(status_code=404, detail=NOTE_NOT_FOUND.format(id=_id))

    return {
        'id': note.id,
        'content': note.content,
        'updated_at': note.updated_at,
    }


@runs_router.put('/{run_id}/note/{_id}')
def update_note_api(run_id, _id: int, note_in: NoteIn, factory=Depends(object_factory)):
    with factory:
        run = factory.find_run(run_id)
        if not run:
            raise HTTPException(status_code=404)

        note = run.find_note(_id=_id)
        if not note:
            raise HTTPException(status_code=404, detail=NOTE_NOT_FOUND.format(id=_id))

        content = note_in.content.strip()
        updated_note = run.update_note(_id=_id, content=content)

    return {
        'id': updated_note.id,
        'content': updated_note.content,
        'updated_at': updated_note.updated_at,
    }


@runs_router.delete('/{run_id}/note/{_id}')
def delete_note_api(run_id, _id: int, factory=Depends(object_factory)):
    with factory:
        run = factory.find_run(run_id)
        if not run:
            raise HTTPException(status_code=404)

        note = run.find_note(_id=_id)
        if not note:
            raise HTTPException(status_code=404, detail=NOTE_NOT_FOUND.format(id=_id))

        run.remove_note(_id)

    return {
        'status': 'OK'
    }


@runs_router.get('/{run_id}/logs/')
async def get_logs_api(run_id: str, record_range: Optional[str] = ''):
    repo = get_project_repo()
    run = get_run_or_404(run_id, repo=repo)

    return StreamingResponse(run_logs_streamer(run, record_range))


def add_api_routes():
    ImageApiConfig.register_endpoints(runs_router)
    TextApiConfig.register_endpoints(runs_router)
    DistributionApiConfig.register_endpoints(runs_router)
    AudioApiConfig.register_endpoints(runs_router)
    FigureApiConfig.register_endpoints(runs_router)



"""empty message

Revision ID: b07e7b07c8ce
Revises: 3c4f22db7a46
Create Date: 2022-04-01 23:49:20.982718

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'b07e7b07c8ce'
down_revision = '3c4f22db7a46'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('note',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('content', sa.Text(), nullable=False),
    sa.Column('run_id', sa.Integer(), nullable=True),
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('updated_at', sa.DateTime(), nullable=True),
    sa.ForeignKeyConstraint(['run_id'], ['run.id'], ),
    sa.PrimaryKeyConstraint('id'),
    )
    op.create_table('note_audit_log',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('note_id', sa.Integer(), nullable=False),
    sa.Column('datetime', sa.DateTime(), nullable=True),
    sa.Column('action', sa.Text(), nullable=True),
    sa.Column('before_edit', sa.Text(), nullable=True),
    sa.Column('after_edit', sa.Text(), nullable=True),
    sa.ForeignKeyConstraint(['note_id'], ['note.id'], ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('id')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('note_audit_log')
    op.drop_table('note')
    # ### end Alembic commands ###
